package examples;

import java.math.BigInteger;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import rns.ECC;
import rns.ECPointRNS;
import rns.Main;
import rns.Rns;

public class Example {

	
	public static void main(String[] args){
		System.out.println("Choose which example to run:");
		System.out.println("1 - Multiplication");
		System.out.println("2 - Multiplication 2");
		System.out.println("3 - Average");
		System.out.println("4 - Modular Multiplication");
		System.out.println("5 - Elliptic Curve Point Scalar Multiplication");
		System.out.println("");
		System.out.print("Type 1 , 2 , 3 , 4 or 5: ");
		
		try {
			int choice;
			if(args.length!=0){
				choice = Integer.valueOf(args[0]);
				System.out.println(choice);
				System.out.println("");
			}else{
				choice = System.in.read();
			}
			switch(choice){
			case 1:
				multExample();
				break;
			case 2:
				multExample2();
				break;
			case 3:
				Average();
				break;
			case 4:
				modularMultExample();
				break;
			default://case 5:
				ECPointAddition();
				break;
			
			}
			
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	
	
	static void multExample(){
//	public static void main(String[] args){	
				
		/*******************************************************************************************************/
		/** Define hardware           											                              **/
		/*******************************************************************************************************/

		/* Define number of bits of the operands and the number of bits in the RNS architecture.
		 * Creates the ModuliSet and hardware structure. Prints hardware configuration to mult.xml.  */
		Main.createModuliSet(64, 16, "mult.xml");
		
		/*Alternatively, moduli set and hardware structure can be loaded from xml file*/
//		Main.loadModuliSet("mult.xml");
		
		/* Creates the output files, appending MULT to their names */
		Main.createFiles("MULT");
		/*******************************************************************************************************/
		
		
		
		/*******************************************************************************************************/
		/** Multiplication Example    											                              **/
		/*******************************************************************************************************/
		
		/* Operands to be set into the data memory, and loaded to the register*/
		Rns A = new Rns("70000","A");        //values are converted to RNS
		Rns B = new Rns("51000","B");        //values are converted to RNS
		
		BigInteger C_bin;	
		Rns C = new Rns("C"); /* reserve register for variable C without setting a value */
		
		
		/* C = A x B */
		A.mult(B, C);
		
		/* Convert back to binary*/
		C_bin = C.toIntMRC();
		
		/*******************************************************************************************************/
		
		
		/* Closes the output files */
		Main.closeFiles();
		
		/* Prints result to Console*/
		System.out.println(C_bin.toString());
	}
		
		
		
		

	static 	void multExample2(){
			
		/*******************************************************************************************************/
		/** Define hardware           											                              **/
		/*******************************************************************************************************/

		/* Define number of bits of the operands and the number of bits in the RNS architecture.
		 * Creates the ModuliSet and hardware structure. Prints hardware configuration to multa.xml.  */
		Main.createModuliSet(64, 16, "multa.xml");
		
		/*Alternatively, moduli set and hardware structure can be loaded from xml file*/
//		Main.loadModuliSet("multa.xml");
		
		/* Creates the output files, appending MULT to their names */
		Main.createFiles("MULTa");
		/*******************************************************************************************************/
		
		
		/*******************************************************************************************************/
		/** Multiplication and Addition w/ Accumulation Example					                              **/
		/*******************************************************************************************************/
		/* Operands */
		Rns A = new Rns("70000","A");
		Rns B = new Rns("51000","B");
		
		BigInteger C_bin;
		Rns C = new Rns("C");	/* reserve register for variable C without setting a value */
		BigInteger D_bin;
		Rns D = new Rns("D");	/* reserve register for variable D without setting a value */
		
		/* C = A x B */
		A.mult(B, C);
		/* D = C + A + B*/
		A.aadd(B, D);
		
		/* Convert back to binary*/
		C_bin = C.toIntMRC();
		D_bin = D.toIntMRC();
		/*******************************************************************************************************/
		
		/* Closes the output files */
		Main.closeFiles();
		
		/* Prints result to Console*/
		System.out.println("C= "+ C_bin.toString());
		System.out.println("D= "+ D_bin.toString());
	}

		
		
		
		
		
	static void Average(){
//	public static void main(String[] args){
			
		
		/*******************************************************************************************************/
		/** Define hardware           											                              **/
		/*******************************************************************************************************/

		/* Define number of bits of the operands and the number of bits in the RNS architecture.
		 * Creates the ModuliSet and hardware structure. Prints hardware configuration to average.xml.  */
		Main.createModuliSet(256, 16, "average.xml");
		
		/*Alternatively, moduli set and hardware structure can be loaded from xml file*/
//		Main.loadModuliSet("average.xml");
		
		/* Creates the output files, appending MULT to their names */
		Main.createFiles("AVERAGE");
		/*******************************************************************************************************/	
		
		
		
		/*******************************************************************************************************/
		/** Average Example	- average of the 3 last inputs						                              **/
		/*******************************************************************************************************/
		
		/* Operands */
		Rns value1 = new Rns(BigInteger.ZERO.toString(),"value1");    //maximum 128 bits
		Rns value2 = new Rns("value2");                               //maximum 128 bits
		Rns value3 = new Rns("value3");                               //maximum 128 bits
		Rns constant = new Rns("33333333333333333333333333333333333333","constant"); // 1/3 * 10^38 (with 126 bits)
		Rns one = new Rns("1", "one");
		
		Rns intermediateResult = new Rns("intermediateResult");
		Rns result = new Rns("result");
		BigInteger result_bin;
		
		
		/* initialize all values to zero */
		value1.mult(one, value2);
		value2.mult(one, value3);
		
		/** for average of the last 3 inputs, loop would start here **/
		Main.codeFile.println("-- loop would start here");
		Main.dataFile.println("-- loop would start here");
		
		/* shift values */
		value2.mult(one, value3);
		value1.mult(one, value2);
		/* read new value */
		value1 = new Rns("45678","value1");
		
		/* add values */
		value1.add(value2, intermediateResult);
		intermediateResult.add(value3, intermediateResult);
		
		/* multiply by 1/3 x 10 ^ 38 */
		intermediateResult.mult(constant, result);
		
		/* Convert back to binary*/
		result_bin = result.toIntMRC();
		
		/*******************************************************************************************************/
		
		
		/* Closes the output files */
		Main.closeFiles();
		
		System.out.println(result_bin.doubleValue()*10e-38);
	}

		
		
		
		
	static void modularMultExample(){
				
		/*******************************************************************************************************/
		/** Define modulo and hardware           											                  **/
		/*******************************************************************************************************/
		
		/* Set the modulo for the modulo operation */
		BigInteger N = new BigInteger("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFFFFFFFFFFFF",16);  /* modulo */
		
		
		/*Define number of bits of the operands and the number of bits in the RNS architecture.
		 *Creates the ModuliSet and hardware structure suitable for Modular operation. 
		 *Prints hardware configuration to mmult.xml. */
		Main.createModuliSetWithMM(256, 16, "mmult.xml",N);
		
		
		/*Alternatively, moduli set and hardware structure can be loaded from xml file*/
//		Main.loadModuliSet("mmult.xml");
	
		/* Creates the output files, appending MMULT to their names 
		 * Must be called after creating or loading ModuliSet*/
		Main.createFiles("MMULT");
		
		/* Generate and write Montgomery Modular Multiplication constants.
		 * Verifies if the current moduliSet is compatible with the selected modulo N*/
		Rns.generateMMConstantsFor(N);
		/*******************************************************************************************************/
		
		
		
		/*******************************************************************************************************/
		/** Modular Multiplication Example										                              **/
		/*******************************************************************************************************/
		
		/* Operands */
		BigInteger A_bin = new BigInteger("FFFFFFFFFFFFFFFFFFFF0000FFFFFFFEFFFFFFFFFFFEFFFE",16);
		BigInteger B_bin = new BigInteger("FFFFFFFFFFFFFFFFFFFFFFFFFFF0000EFFFFFFFFFFFFFFFE",16);
		BigInteger C_bin;
		Rns A = new Rns(A_bin.toString(),"A");      //values are converted to RNS 
		Rns B = new Rns(B_bin.toString(),"B");      //values are converted to RNS 
		Rns C = new Rns("C");
		Rns A_MD = new Rns("A_MD");  /* register allocation to be used in the Montgomery Domain representation */
		Rns B_MD = new Rns("B_MD");  /* register allocation to be used in the Montgomery Domain representation */
		Rns C_MD = new Rns("C_MD");  /* register allocation to be used in the Montgomery Domain representation */
		
		/* Convert operands to the Montgomery Domain */
		A_MD = A.toMD();
		B_MD = B.toMD();
		
		/* C = A x B mod N */
		A_MD.modularMult(B_MD, C_MD);
		
		/* Convert result from the Montgomery Domain */
		C = C_MD.fromMD();
		
		/* Convert back to binary*/
		C_bin = C.toIntMRC();
		
		/*******************************************************************************************************/
		
		
		
		
		/* Closes the output files */
		Main.closeFiles();
		
		System.out.println(C_bin + " which should be equal to \n" + A_bin.multiply(B_bin).mod(N));
	}

	
	
	
	
	static void ECPointAddition(){
		
		
		/*******************************************************************************************************/
		/** Define modulo and hardware           											                  **/
		/*******************************************************************************************************/
		
		/* Set the modulo for the modulo operation and EC parameters */
		BigInteger p = new BigInteger("263");   /* modulo */
		BigInteger a_bin = new BigInteger("2");   /* Curve parameters */
		BigInteger b_bin = new BigInteger("3");   /* Curve parameters */
		
		
		/*Define number of bits of the operands and the number of bits in the RNS architecture.
		 *Creates the ModuliSet and hardware structure suitable for Modular operation. 
		 *Prints hardware configuration to padd.xml. */
		Main.createModuliSetWithMM(256, 16, "padd.xml",p);
		
		
		/*Alternatively, moduli set and hardware structure can be loaded from xml file*/
//		Main.loadModuliSet("padd.xml");

		/* Creates the output files, appending MMULT to their names 
		 * Must be called after creating or loading ModuliSet*/
		Main.createFiles("PADD");
		
		/* Generate and write Montgomery Modular Multiplication EC constants.
		 * Verifies if the current moduliSet is compatible with the selected modulo p*/
		ECPointRNS.generateECconstants(a_bin,b_bin,p);
		/*******************************************************************************************************/
		
		
		
		
		
		/*******************************************************************************************************/
		/** Elliptic Curve Point Addition										                              **/
		/*******************************************************************************************************/
		
		
		/* Set the binary inputs */
		BigInteger X1_bin = new BigInteger("14");
		BigInteger Y1_bin = new BigInteger("113");

		BigInteger X2_bin = new BigInteger("13");
		BigInteger Y2_bin = new BigInteger("139");
		
		BigInteger X3_bin;
		BigInteger Y3_bin;
		BigInteger Z3_bin;

		

		
		Main.codeFile.println("-- Point Add Start");
		Main.dataFile.println("-- Point Add Start");
		
		/* Operands for P3 = P1 + P2*/ /* P(x,y) */
		
		/*Inputs*/
		Rns X1      = new Rns(X1_bin.toString(),"X1");
		Rns Y1      = new Rns(Y1_bin.toString(),"Y1");
		
		Rns X2      = new Rns(X2_bin.toString(),"X2");
		Rns Y2      = new Rns(Y2_bin.toString(),"Y2");
		
		/*Outputs - projective coordinates of P3*/
		Rns X3;
		Rns Y3;
		Rns Z3;
		
		
		
		/*Computation begins*/		
		/*Intermediate values*/
		Rns X1_MD;
		Rns Y1_MD;
		Rns X2_MD;
		Rns Y2_MD;
		
		Rns Z1_MD      = new Rns("Z1_MD");
		Rns Z2_MD      = new Rns("Z2_MD");
		Rns X3_MD = new Rns("X3_MD");
		Rns Y3_MD = new Rns("Y3_MD");
		Rns Z3_MD = new Rns("Z3_MD");
		Rns A       = new Rns("A");
		Rns B       = new Rns("B");
		Rns C       = new Rns("C");
		Rns D       = new Rns("D");
		Rns E       = new Rns("E");
		Rns F       = new Rns("F");
		Rns G       = new Rns("G");
		Rns H       = new Rns("H");
		Rns I       = new Rns("I");
		Rns lambdaN = new Rns("lambdaN");
		Rns lambdaD = new Rns("lambdaD");
		
		
		/* Convert operands to the Montgomery Domain */
		X1_MD = X1.toMD();
		Y1_MD = Y1.toMD();
		X2_MD =  X2.toMD();
		Y2_MD =  Y2.toMD();
		
		Rns.one.mult(Rns.MmodN, Z1_MD);     /* initialize Z projective coordinate to 1, in the Montgomery Domain. Z1 = 1 . M mod p)*/
		Rns.one.mult(Rns.MmodN, Z2_MD);     /* initialize Z projective coordinate to 1, in the Montgomery Domain. Z2 = 1 . M mod p)*/
		
		/* Perform point addition according to unified addition formulae in */
		/* Jean-Claude Bajard, Sylvain Duquesne, Milos D. Ercegovac. COMBINING LEAK
		 * RESISTANT ARITHMETIC FOR ELLIPTIC CURVES DEFINED OVER F p AND RNS
		 * REPRESENTATION. Publications Mathematiques UFR Sciences Techniques Besan con, 2013,
		 * pp.67-87
		 */
		
		/** A = X2.Z1 */
		X2_MD.modularMult(Z1_MD, A);
		
		/** B = X1.Z2 */
		X1_MD.modularMult(Z2_MD, B);
		
		/** C = Z1.Z2 */
		Z1_MD.modularMult(Z2_MD, C);
		
		/** D = a.C */
		ECPointRNS.a_MD.modularMult(C, D);
		
		/** lambdaN = (A + B)^2 - A.B + C.D */
		/*I=A+B*/
		A.add(B, I);
		I.modulo(I);
		
		I.modularMult(I, lambdaN);
		A.modularMult(B, B);
		C.modularMult(D, D);
		
		lambdaN.sub(B, lambdaN);
		lambdaN.modulo(lambdaN);
		lambdaN.add(D, lambdaN);
		lambdaN.modulo(lambdaN);
		
		/** E = Y1.Z2 + Y2.Z1 */
		Y1_MD.modularMult(Z2_MD, E);
		Y2_MD.modularMult(Z1_MD, H);
		E.add(H, E);
		E.modulo(E);
		
		
		/** lambdaD = E.C */
		E.modularMult(C, lambdaD);
		
		/**F = E.lambdaD */
		E.modularMult(lambdaD, F);
		
		/** G = lambdaN.lambdaN */
		lambdaN.modularMult(lambdaN, G);
		
		/** H = F.(A+B) = F.I */
		F.modularMult(I, H);
		
		/** X3 = 2.lambdaD . (G - H) */
		G.sub(H, A);
		A.modulo(A);
		
		A.modularMult(lambdaD, X3_MD);
		X3_MD.mult(ECPointRNS.two, X3_MD);
		X3_MD.modulo(X3_MD);
		
		/** Y3 = lambdaN(3.H - 2.G) - F^2 */
		F.modularMult(F, F);
		G.mult(ECPointRNS.two, G);
		G.modulo(G);
		ECPointRNS.two.add(Rns.one, B); //3
		B.mult(H, H);
		H.modulo(H);
		
		H.sub(G, H);
		H.modulo(H);
		lambdaN.modularMult(H, lambdaN);
		lambdaN.sub(F, Y3_MD);
		Y3_MD.modulo(Y3_MD);
		
		/** Z3 = 2.lambdaD^3 */
		lambdaD.modularMult(lambdaD, A);
		A.modularMult(lambdaD, Z3_MD);
		Z3_MD.mult(ECPointRNS.two, Z3_MD);
		Z3_MD.modulo(Z3_MD);
		
			
		/* Convert result from the Montgomery Domain */
//		Main.codeFile.println("-- Conversion to OD");
		X3 = X3_MD.fromMD();
		Y3 = Y3_MD.fromMD();
		Z3 = Z3_MD.fromMD();
		
		/* Convert back to binary*/
//		Main.codeFile.println("-- Conversion to bin");
		X3_bin = X3.toIntMRC();
		Y3_bin = Y3.toIntMRC();
		Z3_bin = Z3.toIntMRC();
		
		/*******************************************************************************************************/
		
		
		
		/* Closes the output files */
		Main.closeFiles();
		
		/*Binary verification of the result*/
		System.out.println("x = " + X3_bin.multiply(Z3_bin.modInverse(p)).mod(p));
		System.out.println("y = " + Y3_bin.multiply(Z3_bin.modInverse(p)).mod(p));
		
		
		
		
		/**Binary Version*/
/*		BigInteger a = new BigInteger("2");  
		BigInteger b = new BigInteger("3"); 
		BigInteger p = new BigInteger("263"); 
		
		BigInteger X1 = new BigInteger("14");
		BigInteger Y1 = new BigInteger("113");
		BigInteger Z1 = new BigInteger("1");
		
		BigInteger X2 = new BigInteger("13");
		BigInteger Y2 = new BigInteger("139");
		BigInteger Z2 = new BigInteger("1");
		
		BigInteger X3;
		BigInteger Y3;
		BigInteger Z3;
		
		BigInteger lambdaN;
		BigInteger lambdaD;
		BigInteger A;
		BigInteger B;
		BigInteger C;
		BigInteger D;
		BigInteger E;
		BigInteger F;
		BigInteger G;
		BigInteger H;
//		
//		temp
//		lambdaN = X1.multiply(Z2).add(X2.multiply(Z1)).pow(2).subtract(X1.multiply(X2).multiply(Z1.multiply(Z2)).add(a.multiply(Z1.pow(2).multiply(Z2.pow(2))))).mod(p);
//		lambdaD = Y1.multiply(Z2).add(Y2.multiply(Z1)).multiply(Z1).multiply(Z2).mod(p);
//		
//		
//		X3 = new BigInteger("2").multiply(lambdaD).multiply(lambdaN.pow(2).subtract(X1.multiply(X2).add(X2.multiply(Z1)).multiply(Y1.multiply(Z2).add(Y2.multiply(Z1))).multiply(lambdaD))).mod(p);
		
		A = X2.multiply(Z1).mod(p);
		B = X1.multiply(Z2).mod(p);
		C = Z1.multiply(Z2).mod(p);
		D = a.multiply(C).mod(p);
		lambdaN = A.add(B).pow(2).subtract(A.multiply(B)).add(C.multiply(D)).mod(p);
		
		E = Y1.multiply(Z2).add(Y2.multiply(Z1)).mod(p);
		lambdaD = E.multiply(C).mod(p);
		
		F = E.multiply(lambdaD).mod(p);
		G = lambdaN.pow(2).mod(p);
		H = F.multiply(A.add(B)).mod(p);
		
		X3 = new BigInteger("2").multiply(lambdaD).multiply(G.subtract(H)).mod(p);
		Y3 = lambdaN.multiply(new BigInteger("3").multiply(H).subtract(new BigInteger("2").multiply(G))).subtract(F.pow(2)).mod(p);
		Z3 = new BigInteger("2").multiply(lambdaD.pow(3)).mod(p);
		
		System.out.println("x = " + X3.multiply(Z3.modInverse(p)).mod(p));
		System.out.println("y = " + Y3.multiply(Z3.modInverse(p)).mod(p));
		
		System.out.println("A = " + A);
		System.out.println("B = " + B);
		System.out.println("C = " + C);
		System.out.println("D = " + D);
		System.out.println("E = " + E);
		System.out.println("F = " + F);
		System.out.println("G = " + G);
		System.out.println("H = " + H);
		System.out.println("lambdaN = " + lambdaN);
		System.out.println("lambdaD = " + lambdaD);
		System.out.println("X3 = " + X3);
		System.out.println("Y3 = " + Y3);
		System.out.println("Z3 = " + Z3); */
	}
	
	
	

}
