package rns;

import gui.MainWindow;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;

/**
 * Class implementing top Elliptic Curve Cryptography operation
 * @author Juvenal
 *
 */
public class ECC implements Runnable{
	public static BigInteger x;
	public static BigInteger y;
	public static BigInteger a;
	public static BigInteger b;
	public static BigInteger scalar;
	public static BigInteger N;
	
	public void run() {
		try {
			ECC_PointScalarMultiplication();
		} catch (FileNotFoundException | UnsupportedEncodingException e) {
			e.printStackTrace();
		}
	}

	
	public static void ECC_PointScalarMultiplication() throws FileNotFoundException, UnsupportedEncodingException{
		/******************************************************************/
		/**Initializations												 **/
		/******************************************************************/
		
		/* Using KEY_SIZE and BITS_PER_MOD because they were previously defined by the GUI */
		Main.createModuliSetWithMM(Main.KEY_SIZE, Main.BITS_PER_MOD, "ECC_PS.xml",N); 
		
		/* Creates the output files, appending MMULT to their names 
		 * Must be called after creating or loading ModuliSet*/
		Main.createFiles("ECC_PS");
		


		/* Using a,b and N previously defined by the GUI */
		ECPointRNS.generateECconstants(a,b,N);
		
		
		
		/**Optional - Write RNS and ECC parameters to output moduli file*/
		Main.moduliFile.println("ECC modulo: "+N);
		Main.moduliFile.println("ECC scalar: "+scalar);
		Main.moduliFile.println("ECC x: "+x);
		Main.moduliFile.println("ECC y: "+y);
		Main.moduliFile.println("ECC a: "+a);
		Main.moduliFile.println("ECC b: "+b);
		/******************************************************************/
		
		
		


		/******************************************************************/
		/**Beginning of Point Multiplication							  */
		/******************************************************************/
		ECPoint_ result; /* in binary */
		ECPointRNS PointOne;
		
		/**Point to RNS*/
		PointOne = new ECPointRNS(x,y,false); /* Using x and y previously defined by the GUI */

		/**scalar multiplication*/
		result=PointOne.pointMult(scalar);
		/******************************************************************/
		
		
		
		
		
		/******************************************************************/
		/**Setting Results to GUI and output file						  */
		/******************************************************************/
		MainWindow.windowBase.setRestultECC(result.x,result.y);
		Main.moduliFile.println("ECC result x: "+result.x);
		Main.moduliFile.println("ECC result y: "+result.y);
		
		Main.closeFiles();
		/******************************************************************/
		
	}
	

	
}
