package rns;

import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Set;

/**
 * Class implementing the emulated co-processor
 * @author Juvenal
 *
 */
public class Hardware {
	
	static RNSChannel[] HWbase1mod; /*RNS channels of the first base*/ 
	static RNSChannel[] HWbase2mod;	/*RNS channels of the second base*/
//	static RNSChannel HWbaseExtra; /*RNS channels of the extra modulus*/
	static BigInteger outRegister; /*register that emulates the outbus*/
	
	private static Hashtable<String, Integer> registerNames;
	private static Hashtable<Integer, String > opcodes;
	private static int currentRegisterNum;
	private static int maxNumberOfRegisters;
	
	
	public Hardware(){
		HWbase1mod = new RNSChannel[Rns.h1];
		HWbase2mod = new RNSChannel[Rns.h2];
		maxNumberOfRegisters = Rns.maxNumberOfRegisters;
		currentRegisterNum=4;
		registerNames = new Hashtable<String, Integer>();
		opcodes = new Hashtable<Integer, String>(10);
		opcodes.put(0, "add_a");
		opcodes.put(1, "sub_a");
		opcodes.put(2, "mul_a");
		opcodes.put(8, "aadd_a");
		opcodes.put(9, "asub_a");
		opcodes.put(10, "amul_a");
		opcodes.put(12, "sadd_a");
		opcodes.put(13, "ssub_a");
		opcodes.put(14,"smul_a");
		opcodes.put(15,"sync");
		
	}
	
	
	/**
	 * Configures each of the moduli
	 * @param m
	 * 			modulus
	 * @param i
	 * 			position
	 * @param base
	 * 			base of the modulus i
	 */
	void setChannel(BigInteger m, int i, int base) {
		if(base == 1){
			HWbase1mod[i] = new RNSChannel(m,i,base);
		}else{
			if(base==2){
				HWbase2mod[i] = new RNSChannel(m,i,base);
			}
//			else{
//				HWbaseExtra = new RNSChannel(m,i,base);
//			}
		}
	}
	
	
	/**
	 * Creates register inside all RNS channels
	 * @param name
	 * 				Name of the register
	 */
	void createRegister(String name){
		for(int i=0 ; i<Rns.h1 ; i++){
			HWbase1mod[i].addRegister(name, BigInteger.ZERO);
			HWbase2mod[i].addRegister(name, BigInteger.ZERO);
		}
//		HWbaseExtra.addRegister(name, BigInteger.ZERO);
	}
	
	
	/**
	 * Get the number of registers used in a given base
	 * @param base
	 * 				Base to know the number of registers
	 * @return
	 * 				The number of registers
	 */
	public static int getNumberOfRegisters(int base){
		if(base==1){
			return HWbase1mod[Rns.h1-1].registerBank.size();
		}
//		else{
			return HWbase2mod[Rns.h2-1].registerBank.size();
//		}
//		return HWbaseExtra.registerBank.size();
	}
	
	
	/**
	 * MRS based base extension.
	 * Side effect: Uses auxiliary "variable3" in both bases 
	 * @param src
	 * 			Source base
	 * @param dst
	 * 			Destination base
	 * @param srcReg
	 * 			Source register name
	 * @param dstReg
	 * 			Destination register name
	 */
	void baseExtensionMRS(int src,int dst, String srcReg,String dstReg){
		
		 
		
		
		if (src == 1 && dst == 2) {
			instructionNoPrint("100", RNSModule.MUL_VALUE, "variable3", srcReg, "12mrsBE" + 0);
			Hardware.HWbase1mod[0].writeOutRegister("variable3");
			instructionNoPrint("011", RNSModule.MUL_VALUE, "variable3", "OUTBUS", "12mrsBE" + 1);
			instructionNoPrint("100", RNSModule.AMUL_VALUE, "variable3", "OUTBUS", "12mrsBE" + 1);
			Hardware.HWbase1mod[1].writeOutRegister("variable3");
			for (int j = 2; j < Rns.h2; j++) {
				instructionNoPrint("111", RNSModule.AMUL_VALUE, "variable3", "OUTBUS", "12mrsBE" + j);
				Hardware.HWbase1mod[j].writeOutRegister("variable3");
			}
			instructionNoPrint("011", RNSModule.AMUL_VALUE, dstReg, "OUTBUS", "12mrsBE" + Rns.h1);
			Main.assemblyFile.println("-- "+"01" + " " + "RNS-to-MRC_p" + " " + "dummy" + " " + srcReg + " "+ "-----");
			Hardware.printHexInstruction("01", "0110", "1", "dummy", srcReg, null, Main.codeFile);
			Main.assemblyFile.println("-- "+"10" + " " + "MRC-to-RNS" + " " + dstReg + " " + "-----" + " "+ "-----");
			Hardware.printHexInstruction("10", "0101", "0", dstReg, null, null, Main.codeFile);
		}else{
			instructionNoPrint("010", RNSModule.MUL_VALUE, "variable3", srcReg, "21mrsBE" + 0);
			Hardware.HWbase2mod[0].writeOutRegister("variable3");
			instructionNoPrint("101", RNSModule.MUL_VALUE, "variable3", "OUTBUS", "21mrsBE" + 1);
			instructionNoPrint("010", RNSModule.AMUL_VALUE, "variable3", "OUTBUS", "21mrsBE" + 1);
			Hardware.HWbase2mod[1].writeOutRegister("variable3");
			for (int j = 2; j < Rns.h2; j++) {
				instructionNoPrint("111", RNSModule.AMUL_VALUE, "variable3", "OUTBUS", "21mrsBE" + j);
				Hardware.HWbase2mod[j].writeOutRegister("variable3");
			}
			instructionNoPrint("101", RNSModule.AMUL_VALUE, dstReg, "OUTBUS", "21mrsBE" + Rns.h1);
			Main.assemblyFile.println("-- "+"10" + " " + "RNS-to-MRC_p" + " " + "dummy" + " " + srcReg + " "+ "-----");
			Hardware.printHexInstruction("10", "0110", "1", "dummy", srcReg, null, Main.codeFile);
			Main.assemblyFile.println("-- "+"01" + " " + "MRC-to-RNS" + " " + dstReg + " " + "-----" + " "+ "-----");
			Hardware.printHexInstruction("01", "0101", "0", dstReg, null, null, Main.codeFile);
		}
	}
	
	
	/**
	 * Executes instruction on the Hardware and prints it to output file.
	 * @param destination
	 * 					Bases to which the instruction is to
	 * @param opcode
	 * 					Instruction opcode
	 * @param D
	 * 					Name of the destination register
	 * @param A
	 * 					Name of the operand A register
	 * @param B
	 * 					Name of the operand B register
	 */
	void instruction(String destination,int opcode, String D, String A, String B){
		Main.assemblyFile.println("-- "+new StringBuilder(destination).reverse().toString().substring(1) + " " + opcodes.get(opcode) + " " + D + " " + A + " "+ B);
		instructionNoPrint( destination,opcode, D, A, B);
		Hardware.printHexInstruction(new StringBuilder(destination).reverse().toString().substring(1), Integer.toString(opcode+0b10000, 2).substring(1), "1", D, A, B, Main.codeFile);

	}
	
	/**
	 * Executes instruction on the Hardware without printing to the output file
	 * @param destination
	 * 					Bases to which the instruction is to
	 * @param opcode
	 * 					Instruction opcode
	 * @param D
	 * 					Name of the destination register
	 * @param A
	 * 					Name of the operand A register
	 * @param B
	 * 					Name of the operand B register
	 */
	void instructionNoPrint(String destination,int opcode, String D, String A, String B){
		for(int i = 0; i < Rns.h1 ; i++){
			if(destination.charAt(0)=='1'){
				HWbase1mod[i].execute(opcode, D, A, B);
			}
				
			if(destination.charAt(1)=='1'){
				HWbase2mod[i].execute(opcode, D, A, B);
			}
		}

		
	}
	
	/**
	 * Hexadecimal printing in chunks of n bits
	 * @param value
	 * 				Binary value to print
	 * @param n
	 * 				Number of bits for each chunk
	 * @param file
	 * 				Output file to print to
	 */
	static void printHexn(BigInteger value, int n, PrintWriter file){
		for(int i = 0 ; i < Main.MODULI_NO ; i++){
			file.println(String.format("%0"+(int)Math.ceil(Main.BITS_PER_MOD/4.0)+"X", value.and(new BigInteger("2").pow(n).subtract(BigInteger.ONE))));
			value=value.shiftRight(n);
		}
	}
	

	/**
	 * Hexadecimal printing of RNS instructions
	 * @param base
	 * 				Bases to which the instruction is to
	 * @param opcode
	 * 				Instruction opcode
	 * @param a_p
	 * 				Accumulate/propagate
	 * @param rd
	 * 				Name of the destination register
	 * @param rs1
	 * 				Name of the operand A register
	 * @param rs2
	 * 				Name of the operand A register
	 * @param file
	 * 				Output file to print to
	 */
	static void printHexInstruction(String base, String opcode, String a_p, String rd,String rs1,String rs2, PrintWriter file){
		int rd_int=0, rs1_int=0,rs2_int=0;
		
		if(rd!=null && !registerNames.containsKey(rd)){ 
			if(currentRegisterNum<maxNumberOfRegisters){
				System.out.println("R"+currentRegisterNum+": "+ rd);
				registerNames.put(rd, currentRegisterNum++);
			}else{
				System.out.println("Maximum number of registers in the co-processor reached. Aborting...");
				System.exit(-1);
			}
		}else{
			
		}
		
		if(rd!=null)
			rd_int=registerNames.get(rd);
		if(rs1!=null)
			rs1_int=registerNames.get(rs1);
		if(rs2!=null)
			rs2_int=registerNames.get(rs2);
		

		
		file.println(Integer.toHexString(Integer.parseInt(base+opcode+a_p+Integer.toBinaryString(rd_int+0b100000000).substring(1)+Integer.toBinaryString(rs1_int+0b100000000).substring(1)+Integer.toBinaryString(rs2_int+0b100000000).substring(1), 2)));
	}
	
	/**
	 * Attributes a name of a register to a specific Register Number, when printing the generated codeFile
	 * Accepts only values from 0 to 3 (inclusive)
	 * @param name Name of the register
	 * @param num Number to be associated to that register name
	 */
	static void setRegister(String name, int num){
		if(num >=0 && num < 4){
			System.out.println("R"+num+": "+ name);
			registerNames.put(name, num);
		}else{
			System.out.println("Ignoring setRegister " + name + " "+num);
		}
	}

}
