package rns;

import gui.MainWindow;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.util.Iterator;
import java.io.IOException;
import java.io.StringWriter;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.Characters;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;





public class Main {
	public static  int KEY_SIZE = 64; /* number of bits of the key */
	public static  int BITS_PER_MOD = 16; /* number of bits per modulus */
	public static  int MODULI_NO = 3; /* number of moduli per base*/
	public static PrintWriter dataFile;
	public static PrintWriter codeFile;
	public static PrintWriter resultFile;
	public static PrintWriter moduliFile;
	public static PrintWriter assemblyFile;
	static String name;
	
	
	

	static MainWindow windowBase;
	
	/**
	 * Defines number of bits
	 * @param operandSize
	 * 		The size of the biggest operand, in bits
	 * @param bitsPerModuli
	 * 		The number of bits per RNS channel
	 */
	public static void  setBitsInfo(int operandSize,int bitsPerModuli){
		KEY_SIZE = operandSize;
		BITS_PER_MOD =bitsPerModuli;
		MODULI_NO= (int)(Math.ceil(KEY_SIZE/(double)BITS_PER_MOD))+1;
	}

	/**
	 * Creates the output files, appending the desired name for the execution
	 * @param name
	 * 		String to be appended to the file name
	 * @throws FileNotFoundException
	 * @throws UnsupportedEncodingException
	 */
	public static void createFiles(String name){
		try {
			Main.name=name;
			dataFile = new PrintWriter("key"+KEY_SIZE+"-n"+BITS_PER_MOD+"-data"+name+ ".txt", "UTF-8");
			codeFile = new PrintWriter("key"+KEY_SIZE+"-n"+BITS_PER_MOD+"-code"+name+ ".txt", "UTF-8");
			resultFile = new PrintWriter("key"+KEY_SIZE+"-n"+BITS_PER_MOD+"-result"+name+ ".txt", "UTF-8");
			moduliFile = new PrintWriter("key"+KEY_SIZE+"-n"+BITS_PER_MOD+"-moduli"+name+ ".txt", "UTF-8");
			assemblyFile=new PrintWriter("key"+KEY_SIZE+"-n"+BITS_PER_MOD+"-assembly"+name+ ".txt", "UTF-8");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}
	}
	
	public static void closeFiles(){
		dataFile.close();
		codeFile.close();
		resultFile.close();
		moduliFile.close();
		assemblyFile.close();
	}
	
	static void printModuli(){
		Main.moduliFile.println("B1                    B2");
		for (int i=0;i< Rns.h1;i++){
			Main.moduliFile.println(Rns.base1mod[i] + "(k=" + new BigInteger("2").pow(Main.BITS_PER_MOD).subtract(Rns.base1mod[i]) + ")   " + Rns.base2mod[i] + "(k="
					+ new BigInteger("2").pow(Main.BITS_PER_MOD).subtract(Rns.base2mod[i]) +  ")");
		}
	}
	
	
	
	public static void loadModuliSet(String configurationFile){
		
		XMLReader reader = new XMLReader(configurationFile);
		String attributes;
		
		Rns.maxNumberOfRegisters = Integer.valueOf(reader.getValueOf("maxNumberOfRegisters"));
		KEY_SIZE = Integer.valueOf(reader.getValueOf("dynamicRange_nBits"));
		BITS_PER_MOD = Integer.valueOf(reader.getValueOf("channelLenght"));
		int numberOfBases = Integer.valueOf(reader.getAttributesOf("bases").split(" ")[0]);
		MODULI_NO = Integer.valueOf(reader.getAttributesOf("moduliSet").split(" ")[0]);
		
		
		
		Rns.n = Main.BITS_PER_MOD;
		Rns.h1 = Main.MODULI_NO;
		Rns.h2 = Rns.h1;
//		KEY_SIZE= (Rns.h1-1)*BITS_PER_MOD;
		
		Rns.base1mod = new BigInteger[Rns.h1];
		Rns.base2mod = new BigInteger[Rns.h2];
		Rns.HW = new Hardware();	
		
		for(int i = 0;i<Rns.h1;i++){
			Rns.base1mod[i]=new BigInteger("2").pow(BITS_PER_MOD).subtract(new BigInteger(reader.getValueOf("k")));
			Rns.HW.setChannel(Rns.base1mod[i], i, 1);
		}
		reader.getEndElement("moduliSet");
		reader.getStartElement("moduliSet");
		for(int i = 0;i<Rns.h2;i++){
			Rns.base2mod[i]=new BigInteger("2").pow(BITS_PER_MOD).subtract(new BigInteger(reader.getValueOf("k")));
			Rns.HW.setChannel(Rns.base2mod[i], i, 2);
		}
		reader.getEndElement("moduliSet");
		reader.getEndElement("bases");
		
		/* test for /RNS  */
		if(reader.isEndElement()){
			Rns.generateConstants();
		}else{
			/*constants*/
			reader.getStartElement("constants");
			while(!reader.isEndElement()){
				attributes=reader.getAttributesOf("constant");
				reader.getStartElement("base");
				for(int i=0 ; i < Rns.h1 ; i++){
					Hardware.HWbase1mod[i].addRegister(attributes.split(" ")[0],new BigInteger(reader.getValueOf("ch"+i)));
				}
				reader.getEndElement("base");
				if(Integer.valueOf(attributes.split(" ")[1])==2){
					reader.getStartElement("base");
					for(int i=0 ; i < Rns.h2 ; i++){
						Hardware.HWbase2mod[i].addRegister(attributes.split(" ")[0],new BigInteger(reader.getValueOf("ch"+i)));
					}
					reader.getEndElement("base");
				}
				reader.getEndElement("constant");
			}
			reader.getEndElement("constants");
		
		}
				

	}		
		
	
	/**
	 * For use when NO Modular Multiplication is needed.
	 * Defines the number of bits of the operands and the number of bits in the RNS architecture. 
	 * Configures the RNS channels and hardware.
	 * Prints Hardware configuration to output file.
	 * @param operandSize The size of the biggest operand, in bits
	 * @param bitsPerModulus The number of bits per RNS channel
	 * @param outputFilename Name of the xml file to write the Hardware Configuration
	 */
	public static void createModuliSet(int operandSize, int bitsPerModulus, String outputFilename){
		BigInteger biggestNumber = new BigInteger("2").pow(operandSize).subtract(BigInteger.ONE); /*Representing the biggest integer to be used*/
		createModuliSetWithMM(operandSize,bitsPerModulus,outputFilename,biggestNumber);
	}
	
	
	/**
	 * For use when Modular Multiplication is needed.
	 * Defines the number of bits of the operands and the number of bits in the RNS architecture. 
	 * Configures the RNS channels and hardware accordingly to the MM modulo.
	 * Prints Hardware configuration to output file.
	 * @param operandSize The size of the biggest operand, in bits
	 * @param bitsPerModulus The number of bits per RNS channel
	 * @param outputFilename Name of the xml file to write the Hardware Configuration to
	 * @param modulo Modulo for the Modular Multiplication
	 */
	public static void createModuliSetWithMM(int operandSize, int bitsPerModulus, String outputFilename,BigInteger modulo){
		
		setBitsInfo(operandSize, bitsPerModulus);

		Rns.computeModuli(modulo);
		
        Rns.generateConstants();
		
		try {
	         XMLOutputFactory xMLOutputFactory = XMLOutputFactory.newInstance();	
	         XMLStreamWriter xmlFile = xMLOutputFactory.createXMLStreamWriter(new FileWriter(outputFilename));
	   
	         xmlFile.writeStartDocument();
	         xmlFile.writeStartElement("RNS");
	   
	         xmlFile.writeStartElement("maxNumberOfRegisters");			
	         xmlFile.writeCharacters(String.valueOf(Rns.maxNumberOfRegisters));
	         xmlFile.writeEndElement();
	         
	         xmlFile.writeStartElement("dynamicRange_nBits");			
	         xmlFile.writeCharacters(String.valueOf(operandSize));
	         xmlFile.writeEndElement();
	         
	         xmlFile.writeStartElement("channelLenght");			
	         xmlFile.writeCharacters(String.valueOf(bitsPerModulus));
	         xmlFile.writeEndElement();
	         
	         
	         
	         xmlFile.writeStartElement("bases");
	         xmlFile.writeAttribute("numberOfbases", "2");
	         

	         
	         
	         
	         xmlFile.writeStartElement("moduliSet");
	         xmlFile.writeAttribute("numberOfModuli", String.valueOf(Main.MODULI_NO));
	         for (int i=0;i< Rns.h1;i++){
	        	 xmlFile.writeStartElement("k");
	        	 xmlFile.writeCharacters(new BigInteger("2").pow(bitsPerModulus).subtract(Rns.base1mod[i]).toString());
	        	 xmlFile.writeEndElement();
	         }
	         xmlFile.writeEndElement();
	         
	         xmlFile.writeStartElement("moduliSet");
	         xmlFile.writeAttribute("numberOfModuli", String.valueOf(Main.MODULI_NO));
	         for (int i=0;i< Rns.h2;i++){
	        	 xmlFile.writeStartElement("k");
	        	 xmlFile.writeCharacters(new BigInteger("2").pow(bitsPerModulus).subtract(Rns.base2mod[i]).toString());
	        	 xmlFile.writeEndElement();
	         }
	         xmlFile.writeEndElement();
	         
	         /*end bases*/
	         xmlFile.writeEndElement();


	         
	         
	         
	         xmlFile.writeStartElement("constants");
	         /**
	 		 * Forward Conversion
	 		 */
	         for (int j = 1; j <= Math.ceil(Main.KEY_SIZE / (float) Rns.n); j++) {
	        	 xmlFile.writeStartElement("constant");
	        	 xmlFile.writeAttribute("name", "2^(n*j-1)"+j);
		         xmlFile.writeAttribute("numberOfBases", "2");
	        	 
	        	 xmlFile.writeStartElement("base");
	        	 for (int i = 0; i < Rns.h1 ; i++){
	        		 xmlFile.writeStartElement("ch"+i);
	        		 xmlFile.writeCharacters(Hardware.HWbase1mod[i].R("2^(n*j-1)" + j).toString());
	        		 xmlFile.writeEndElement();
	        	 }
	        	 xmlFile.writeEndElement();
	        	 
	        	 xmlFile.writeStartElement("base");
	        	 for (int i = 0; i < Rns.h1 ; i++){
	        		 xmlFile.writeStartElement("ch"+i);
	        		 xmlFile.writeCharacters(Hardware.HWbase2mod[i].R("2^(n*j-1)" + j).toString());
	        		 xmlFile.writeEndElement();
	        	 }
	        	 xmlFile.writeEndElement();
	        	 
	        	 xmlFile.writeEndElement();
        	 }
	         
	         
	         /**
	          * MRC
	          */
	         for (int j = 0; j < Rns.h1; j++) {
	        	 xmlFile.writeStartElement("constant");
	        	 xmlFile.writeAttribute("name", "newMRCC"+j);
		         xmlFile.writeAttribute("numberOfBases", "1");
		         
		         xmlFile.writeStartElement("base");
	        	 for (int i = 0; i < Rns.h1 ; i++){
	        		 xmlFile.writeStartElement("ch"+i);
	        		 xmlFile.writeCharacters(Hardware.HWbase1mod[i].R("newMRCC" + j).toString());
	        		 xmlFile.writeEndElement();
	        	 }
	        	 xmlFile.writeEndElement();
	        	 
	        	 xmlFile.writeEndElement();
	         }
	         
	         
	         /**
	          * Reverse Conversion
	          */
	         for (int j = 0; j < Rns.h1; j++) {
	        	 xmlFile.writeStartElement("constant");
	        	 xmlFile.writeAttribute("name", "Wij"+j);
		         xmlFile.writeAttribute("numberOfbases", "1");
		         
		         xmlFile.writeStartElement("base");
	        	 for (int i = 0; i < Rns.h1 ; i++){
	        		 xmlFile.writeStartElement("ch"+i);
	        		 xmlFile.writeCharacters(Hardware.HWbase1mod[i].R("Wij" + j).toString());
	        		 xmlFile.writeEndElement();
	        	 }
	        	 xmlFile.writeEndElement();
	        	 
	        	 xmlFile.writeEndElement();
	         }
	         
	         
	         /**
	          * BE
	          */
	         for (int j = 0; j <= Rns.h1; j++) {
	        	 xmlFile.writeStartElement("constant");
	        	 xmlFile.writeAttribute("name", "12mrsBE"+j);
		         xmlFile.writeAttribute("numberOfbases", "2");
	        	 
	        	 xmlFile.writeStartElement("base");
	        	 for (int i = 0; i < Rns.h1 ; i++){
	        		 xmlFile.writeStartElement("ch"+i);
	        		 xmlFile.writeCharacters(Hardware.HWbase1mod[i].R("12mrsBE" + j).toString());
	        		 xmlFile.writeEndElement();
	        	 }
	        	 xmlFile.writeEndElement();
	        	 
	        	 xmlFile.writeStartElement("base");
	        	 for (int i = 0; i < Rns.h1 ; i++){
	        		 xmlFile.writeStartElement("ch"+i);
	        		 xmlFile.writeCharacters(Hardware.HWbase2mod[i].R("12mrsBE" + j).toString());
	        		 xmlFile.writeEndElement();
	        	 }
	        	 xmlFile.writeEndElement();
	        	 
	        	 xmlFile.writeEndElement();
        	 }
	         for (int j = 0; j <= Rns.h1; j++) {
	        	 xmlFile.writeStartElement("constant");
	        	 xmlFile.writeAttribute("name", "21mrsBE"+j);
		         xmlFile.writeAttribute("numberOfbases", "2");
	        	 
	        	 xmlFile.writeStartElement("base");
	        	 for (int i = 0; i < Rns.h1 ; i++){
	        		 xmlFile.writeStartElement("ch"+i);
	        		 xmlFile.writeCharacters(Hardware.HWbase1mod[i].R("21mrsBE" + j).toString());
	        		 xmlFile.writeEndElement();
	        	 }
	        	 xmlFile.writeEndElement();
	        	 
	        	 xmlFile.writeStartElement("base");
	        	 for (int i = 0; i < Rns.h1 ; i++){
	        		 xmlFile.writeStartElement("ch"+i);
	        		 xmlFile.writeCharacters(Hardware.HWbase2mod[i].R("21mrsBE" + j).toString());
	        		 xmlFile.writeEndElement();
	        	 }
	        	 xmlFile.writeEndElement();
	        	 
	        	 xmlFile.writeEndElement();
        	 }
	         
	         
	         
	         
	         /* end constants*/
	         xmlFile.writeEndElement();
	         
	         
	         /*end RNS*/
	         xmlFile.writeEndElement();
	         xmlFile.writeEndDocument();

	         xmlFile.flush();
	         xmlFile.close();




		
		
		
		
	      } catch (XMLStreamException e) {
	         e.printStackTrace();
	      } catch (IOException e) {
	         e.printStackTrace();
	      }
		
	}
	
	
}

