package rns;

import gui.MainWindow;

import java.math.BigInteger;
import java.util.Hashtable;

/**
 * Class implementing the emulated RNS channel
 * @author Juvenal
 *
 */
public class RNSChannel {
	RNSModule hardwareBlock; /* RNS ALU*/
	public Hashtable<String, BigInteger> registerBank;
	private int moduli;
	
	public RNSChannel(BigInteger modulus,int moduli,int base){
		hardwareBlock = new RNSModule(modulus, base);
		registerBank = new Hashtable<String, BigInteger>();
		this.moduli = moduli;
	}
	
	/**
	 * Create or Write to a register in the register bank
	 * @param name
	 * 			Name of the Register
	 * @param value
	 * 			Value to write
	 */
	void addRegister(String name, BigInteger value){
		registerBank.put(name, value);
	}
	
	
	/**
	 * Read from the register bank
	 * @param name
	 * 			Name of the register to be read
	 * @return
	 * 			The stored value
	 */
	BigInteger R(String name){
		return registerBank.get(name);
	}
	
	
	/**
	 * Writes value in outbus. Prints to GUI.
	 * @param value
	 * 			Value to write
	 */
	void writeOutRegister(String value){
		Hardware.outRegister=R(value);
		printInstruction(RNSModule.ADD_VALUE,"OUTBUS",value,"zero");
	}
	
	
	/**
	 * Sends instruction to the ALU. Prints to GUI.
	 * @param opcode
	 * 					Instruction opcode
	 * @param D
	 * 					Name of the destination register
	 * @param A
	 * 					Name of the operand A register
	 * @param B
	 * 					Name of the operand B register
	 */
	void execute(int opcode, String D, String A, String B){
		if(A.equals("OUTBUS")){
			registerBank.put(D, hardwareBlock.execute(opcode, Hardware.outRegister, R(B)));
		}else{
			if(B.equals("OUTBUS")){
				registerBank.put(D, hardwareBlock.execute(opcode, R(A),Hardware.outRegister ));
			}else{
				registerBank.put(D, hardwareBlock.execute(opcode, R(A), R(B)));
			}
		}
		printInstruction(opcode,D,A,B);
	}
	
	/**
	 * Prints instruction to GUI.
	 * @param opcode
	 * 				Instruction opcode
	 * @param result
	 * 				Name of the destination register
	 * @param a
	 * 				Name of the operand A register
	 * @param b
	 * 				Name of the operand B register
	 */
	private void printInstruction(int opcode,String result, String a, String b){
		StringBuilder stringBuilder = new StringBuilder();
		switch (opcode){
			case RNSModule.ADD_VALUE: stringBuilder.append("ADD\n");
							break;
		
			case RNSModule.SUB_VALUE: stringBuilder.append("SUB\n");
							break;
						
			case RNSModule.MUL_VALUE: stringBuilder.append("MUL\n");
							break;
		
			case RNSModule.AADD_VALUE: stringBuilder.append("AADD\n");
							 break;
		
			case RNSModule.ASUB_VALUE: stringBuilder.append("ASUB\n");
							 break;
		
			case RNSModule.AMUL_VALUE: stringBuilder.append("AMUL\n");
							 break;
		
			case RNSModule.SADD_VALUE: stringBuilder.append("SADD\n");
							 break;
		
			case RNSModule.SSUB_VALUE: stringBuilder.append("SSUB\n");
							 break;
		
			case RNSModule.SMUL_VALUE: stringBuilder.append("SMUL\n");
							 break;
		}
		stringBuilder.append(result+"\n"+a+"\n"+b);//+" "+acc);
		
		if(MainWindow.windowBase!=null)
				MainWindow.windowBase.print(stringBuilder.toString(),moduli, hardwareBlock.base);
	}
	
	
	
}
