package rns;

import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;

/**
 * Class implementing the ALU for the RNS channels
 * @author Juvenal
 *
 */
public class RNSModule {
	BigInteger mod;
	int base;
	BigInteger acc;
	BigInteger bin_product;
	
//	static BigInteger last_A=new BigInteger("-1");
//	static BigInteger last_B=new BigInteger("-1");
//	static BigInteger last_result=new BigInteger("-1");
//	static PrintWriter writer;

	static final int ADD_VALUE = 0;
	static final int SUB_VALUE = 1;
	static final int MUL_VALUE = 2;
	static final int AADD_VALUE = 8;
	static final int ASUB_VALUE = 9;
	static final int AMUL_VALUE = 10;
	static final int SADD_VALUE = 12;
	static final int SSUB_VALUE = 13;
	static final int SMUL_VALUE = 14;
	static final int SYNC_VALUE = 15;
	
	
	public RNSModule(BigInteger mod, int base) {
		this.mod = mod;
		this.base = base;
		
//		if(base==2){
//			try {
//				writer = new PrintWriter("debug.txt", "UTF-8");
//			} catch (FileNotFoundException e) {
//				e.printStackTrace();
//			} catch (UnsupportedEncodingException e) {
//				e.printStackTrace();
//			}
//		}
	}
	
	
	/**
	 * Executes instruction according to the ALU architecture
	 * @param opcode
	 * 			Instruction opcode
	 * @param a
	 * 			First operand
	 * @param b
	 * 			Second operand
	 * @return
	 * 			Result
	 */
	BigInteger execute(int opcode, BigInteger a, BigInteger b/*, BigInteger acc*/){
		
		
		
		BigInteger result = new BigInteger("0");
		BigInteger softwareResult = new BigInteger("0");
		BigInteger partial0;
		BigInteger partial1;
		BigInteger twoPowerK=new BigInteger("0");
		BigInteger k;
		BigInteger add_in0 = new BigInteger("0");
		BigInteger add_in1 = new BigInteger("0");
		BigInteger add_in2 = new BigInteger("0");
		BigInteger add_in3 = new BigInteger("0");
		BigInteger add_in4 = new BigInteger("0");
		BigInteger mask_nMinus1Downto0 = new BigInteger("0");
		
		for(int i = 0 ; i<(mod.bitLength());i++){
			mask_nMinus1Downto0=mask_nMinus1Downto0.setBit(i);
		}
		
		
		twoPowerK=twoPowerK.setBit(mod.bitLength());
		k=twoPowerK.subtract(mod);
		
		bin_product = a.multiply(b);
		partial0 = k.multiply(bin_product.shiftRight(Main.BITS_PER_MOD));
		partial1 = k.multiply(partial0.shiftRight(Main.BITS_PER_MOD));
		for ( int i = 2*k.bitLength() ; i < partial1.bitLength() ; i++){
			partial1=partial1.clearBit(i);
		}
		
		add_in2 = b;
		add_in3 = a;
		
		
		switch (opcode){
			case ADD_VALUE: softwareResult = modularSum(a,b);
							break;
							
			case SUB_VALUE: add_in1 = new BigInteger("1").subtract(k);
							add_in2 = b.not().and(mask_nMinus1Downto0);
							softwareResult = modularSub(a,b);
							break;
							
			case MUL_VALUE: add_in2 = bin_product.and(mask_nMinus1Downto0);
							add_in3 = partial0.and(mask_nMinus1Downto0);
							add_in4 = partial1;
							softwareResult = modularMult(a,b);
							break;
							
			case AADD_VALUE: add_in0 = acc;
							 softwareResult = modularSum(modularSum(a,b),acc);
							 break;
							
			case ASUB_VALUE: add_in0 = acc;
							 add_in1 = new BigInteger("1").subtract(k);
							 add_in2 = b.not().and(mask_nMinus1Downto0);
							 softwareResult = modularSum(modularSub(a,b),acc);
							 break;
							
			case AMUL_VALUE: add_in0 = acc;
							 add_in2 = bin_product.and(mask_nMinus1Downto0);
							 add_in3 = partial0.and(mask_nMinus1Downto0);
							 add_in4 = partial1;
							 softwareResult = modularSum(modularMult(a,b),acc);
							 break;
							
			case SADD_VALUE: add_in0 = acc;
							 add_in1 = new BigInteger("2").multiply(new BigInteger("1").subtract(k));
							 add_in2 = b.not().and(mask_nMinus1Downto0);
							 add_in3 = a.not().and(mask_nMinus1Downto0);
							 softwareResult = modularSub(acc,modularSum(a,b));
							 break;
							
			case SSUB_VALUE: add_in0 = acc;
							 add_in1 = new BigInteger("1").subtract(k);
							 add_in3 = a.not().and(mask_nMinus1Downto0);
							 softwareResult = modularSub(acc,modularSub(a,b));
							 break;
							
			case SMUL_VALUE: add_in0 = acc;
			 				 add_in1 = new BigInteger("3").multiply(new BigInteger("1").subtract(k));
			 				 add_in2 = bin_product.not().and(mask_nMinus1Downto0);
			 				 add_in3 = partial0.not().and(mask_nMinus1Downto0);
			 				 add_in4 = partial1.not().and(mask_nMinus1Downto0);
			 				 softwareResult = modularSub(acc,modularMult(a,b));
							 break;
		}
		
		result = modularSum5to1(add_in0,add_in1,add_in2,add_in3,add_in4);
		
		if(result.compareTo(softwareResult) != 0){
			System.out.println("PROBLEM WITH HARDWARE " );
			result = modularSum5to1(add_in0,add_in1,add_in2,add_in3,add_in4);
		}
		acc=result;
		
//		if(base==2 && last_result.compareTo(result)!=0){
//			writer.println(result);
//			last_result=result;
//		}
		
		return result;
	}
	
	
	/**
	 * Modular adder 5:1
	 * @param add_in0
	 * 			First operand
	 * @param add_in1
	 * 			Second operand
	 * @param add_in2
	 * 			Third operand
	 * @param add_in3
	 * 			Fourth operand
	 * @param add_in4
	 * 			Fift operand
	 * @return
	 * 			Result of the modular sum of the 5 operands
	 */
	private BigInteger modularSum5to1(BigInteger add_in0 , BigInteger add_in1 , BigInteger add_in2 , BigInteger add_in3 , BigInteger add_in4/*, BigInteger modulus*/){
		BigInteger result;
		/* values to be added must be < modulus */
		result = modularSum(add_in0.mod(mod),add_in1.mod(mod) );
		result = modularSum(result,add_in2.mod(mod) );
		result = modularSum(result,add_in3.mod(mod) );
		result = modularSum(result,add_in4.mod(mod) );
		
		return result;
	}
	
	
	/**
	 * Binary modular sum
	 * @param a
	 * 			First operand
	 * @param b
	 * 			Second operand
	 * @return
	 * 			Result of the modular sum
	 */
	private BigInteger modularSum(BigInteger a, BigInteger b){
		BigInteger r = new BigInteger("0");
		BigInteger rPrime = new BigInteger("0");
		BigInteger mask = new BigInteger("0");
		BigInteger twoPowerK = new BigInteger("0");
		BigInteger c = new BigInteger("0");
		
		twoPowerK=twoPowerK.setBit(mod.bitLength());
		c=twoPowerK.subtract(mod);

		r = a.add(b);  /* r = a + b */
		rPrime=r.add(c); /*r' = r + p */
		
		if(rPrime.bitLength() >= (mod.bitLength()+1)){
			r= rPrime;
		}
		for(int i = 0 ; i<(mod.bitLength());i++){
			mask=mask.setBit(i);
		}
		return r.and(mask);
	}
	
	
	/**
	 * Binary modular subtraction
	 * @param a
	 * 			First operand
	 * @param b
	 * 			Second operand
	 * @return
	 * 			Result of the modular subtraction
	 */
	private  BigInteger modularSub(BigInteger a, BigInteger b){
		BigInteger r = new BigInteger("0");
		BigInteger rPrime = new BigInteger("0");
		BigInteger mask = new BigInteger("0");
		BigInteger twoPowerK = new BigInteger("0");
		BigInteger c = new BigInteger("0");
		
		twoPowerK=twoPowerK.setBit(mod.bitLength());
		c=twoPowerK.subtract(mod);

		r = a.subtract(b);  /* r = a - b */
		rPrime=r.subtract(c); /*r' = r - p */
		if(r.signum()==-1){
			r= rPrime;
		}
		
		for(int i = 0 ; i<(mod.bitLength());i++){
			mask=mask.setBit(i);
		}
		return r.and(mask);
	}
	
	/**
	 * Binary modular multiplication
	 * @param a
	 * 			First operand
	 * @param b
	 * 			Second operand
	 * @return
	 * 			Result of the modular multiplication
	 */
	private  BigInteger modularMult(BigInteger a, BigInteger b){
		BigInteger r = new BigInteger("0");
		BigInteger rPrime = new BigInteger("0");
		BigInteger mask = new BigInteger("0");
		BigInteger twoPowerK = new BigInteger("0");
		BigInteger c = new BigInteger("0");
		BigInteger aux = new BigInteger("0");
		
		twoPowerK=twoPowerK.setBit(mod.bitLength());
		c=twoPowerK.subtract(mod);
		
		for(int i = 0 ; i<(mod.bitLength());i++){
			mask=mask.setBit(i);
		}

		r = a.multiply(b); /* r=ab */
		while(r.compareTo((mod.add(mod))) >=0){  /* while r >= 2p */
			aux = r.shiftRight(mod.bitLength());	/* <r>_k,1 */
			aux = aux.multiply(c);	/* <r>_k,1 * c */
			r = aux.add(r.and(mask)); /* r = <r>_k,1 * c + <r>_k,0 */
		}
		rPrime = r.add(c);
		
		if(rPrime.bitLength() >= (mod.bitLength()+1)){
			r= rPrime;
		}
		
		
		return r.and(mask);
	}
	
}
