package rns;

import gui.MainWindow;
import java.io.FileNotFoundException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.math.BigInteger;
import java.util.Random;

/**
 * Class implementing top RSA operation
 * @author Juvenal
 *
 */
public class RSA  implements Runnable {
	public static BigInteger MSG;
	public static BigInteger key;
	public static BigInteger publickey;
	public static BigInteger N;	
	
//	private static Rns one;
//	private static Rns MMmodN;
	
	public void run() {
		try {
			RSA_run();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}
	}
	
	
	


	
	private static void RSA_run() throws FileNotFoundException, UnsupportedEncodingException{
		
		/******************************************************************/
		/**Initializations												 **/
		/******************************************************************/
		
		/* Using KEY_SIZE and BITS_PER_MOD because they were previously defined by the GUI */
		Main.createModuliSetWithMM(Main.KEY_SIZE, Main.BITS_PER_MOD, "RSA.xml",N); 
		
		/* Creates the output files, appending MMULT to their names 
		 * Must be called after creating or loading ModuliSet*/
		Main.createFiles("RSA");
		


		/* Using N previously defined by the GUI */
		Rns.generateMMConstantsFor(N);
		
		
		
		/**Optional - Write RSA parameters to output moduli file*/
		Main.moduliFile.println("RSA modulo: "+N);
		Main.moduliFile.println("RSA private key: "+key);
		Main.moduliFile.println("RSA public key: "+publickey);
		Main.moduliFile.println("RSA MSG: "+MSG);
		/******************************************************************/

		Rns value1;
		Rns M_value1 = new Rns("x1");
		
		BigInteger result;
		Rns result_rns = new Rns("result_rns");
		
		
		
		MainWindow.windowBase.setPriting(true);

		
		/**Binary to RNS*/
		value1 = new Rns(MSG.toString(),"x1"); //MSG previously configured by the GUI

		
		/**to Montgomery Domain**/
		value1.modularMult(Rns.MMmodN, M_value1);
		
		
		/**RSA exponentiation**/
		result_rns=M_value1.exp(key); //key previously configured by the GUI
		
		/**from Montgomery Domain to Original Domain**/
		result_rns.modularMult(Rns.one,result_rns);
		
		/**to binary**/
		result = result_rns.toIntMRC();

		
		System.out.println("exp mod N: " + result);
		System.out.println("Expected value: "+MSG.modPow(key, N));


		if(result.compareTo(MSG.modPow(key, N)) != 0){
			System.out.println("/!\\/!\\/!\\/!\\ALERT!!! /!\\/!\\/!\\/!\\ ");
			System.out.println("Values differ!");
			return;
		}
		
		
		
		/**Print result to GUI**/
		MainWindow.windowBase.setRestultRSA(result);
		
		Main.closeFiles();
	}
	
	
	/**
	 * Random generation of RSA key pairs and modulus
	 * @param bits
	 * 			 number of bits of the generated key
	 * @return BigInteger vector where position 0 is the modulus, 1 is the private key and 2 is the public key
	 */
	public static BigInteger[] generateKeys(int bits){
		Random generator = new Random();
		BigInteger p = BigInteger.probablePrime(bits/2,generator);
		BigInteger q = BigInteger.probablePrime(bits/2,generator);
		BigInteger one = new BigInteger("1");
		BigInteger phi;
		BigInteger pub;
		BigInteger priv;
		BigInteger modulus;
		BigInteger[] result = new BigInteger[3];
		
		while(!p.isProbablePrime(8) || p.bitLength()!=bits/2){
			p= new BigInteger(bits/2,generator);
		}
		while(!q.isProbablePrime(8) || q.bitLength()!=bits/2){
			q= new BigInteger(bits/2,generator);
		}
		
		phi = p.subtract(one).multiply(q.subtract(one));
		
		pub = new BigInteger("65537");
		
		modulus = p.multiply(q);
		
		priv = pub.modInverse(phi);
		
		
		
		if(modulus.bitLength()!=bits || priv.bitLength()!=bits){
			return generateKeys(bits);
		}else{
		
			System.out.println("modulus: "+modulus );
			System.out.println("pub: "+pub);
			System.out.println("priv: "+priv);
			result[0]=modulus;
			result[1]=priv;
			result[2]=pub;
			return result;
		}
		
	}
}
