package rns;

import java.math.BigInteger;
import javax.xml.stream.XMLStreamWriter;

/**
 * Class implementing a RNS point. The point is automatically added to the emulated Hardware.
 * @author Juvenal
 *
 */
public class Rns {
	static int n; /* number of bits per channel */
	static int h1;
	static int h2;
	static int maxNumberOfRegisters = 128;
	static BigInteger modulo; /* high level modulo */
	static BigInteger[] base1mod;
	static BigInteger[] base2mod;
//	static BigInteger baseExtra;
	String name;
	static Hardware HW;
	public static Rns one;
	public static Rns MMmodN;
	public static Rns MmodN;
	
	static private String allBases = new String("110");
	static private String firstBase = new String("100");
	static private String secondBase = new String("010");
	static private String noBase = new String("000");
	
	
	

	/**
	 * Creates a register in each RNS channel with no value attributed
	 * @param name
	 * 		name for the register
	 */
	public Rns(String name) {
		this.name = name;

	}
	
	



	/**
	 * Creates a register in each RNS channel with the RNS representation of the binary input value.
	 * @param value
	 * 		String containing the value to be converted to RNS and assigned to the registers
	 * @param name
	 * 		Name of the register
	 */
	public Rns(String value, String name) {
		BigInteger mask = new BigInteger("0");
		BigInteger aux = new BigInteger("0");
		BigInteger aux2;
		BigInteger num = new BigInteger(value);

		this.name = name;
		
		HW.createRegister(name);

		for (int i = 0; i < n; i++) {
			mask = mask.setBit(i);
		}

		/** Conversion from binary to RNS */
		aux2 = num;
		for (int j = 1; j <= Math.ceil(num.bitLength() / (float) n); j++) {
			Hardware.outRegister = aux2.and(mask);
			if (j == 1) {
				HW.instructionNoPrint(allBases, RNSModule.MUL_VALUE, name, "OUTBUS", "2^(n*j-1)" + j);
			} else {
				HW.instructionNoPrint(allBases, RNSModule.AMUL_VALUE, name, "OUTBUS", "2^(n*j-1)" + j);
			}
			aux2 = aux2.shiftRight(n);
		}
		Main.assemblyFile.println("-- "+"01" + " " + "Bin-to-RNS_p" + " " + name + " " + "-----" + " "+ "-----");
		Hardware.printHexInstruction("01","0100","1",name,null,null, Main.codeFile);
		Main.assemblyFile.println("-- "+"10" + " " + "Bin-to-RNS" + " " + name + " " + "-----" + " "+ "-----");
		Hardware.printHexInstruction("10","0100","0",name,null,null, Main.codeFile);
		Hardware.printHexn(num,Main.BITS_PER_MOD,Main.dataFile);
	}

	/**
	 * RNS addition
	 * @param b
	 * 		RNS value to add to
	 * @param destination
	 * 		destination RNS register
	 */
	public void add(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.ADD_VALUE, destination.name, this.name, b.name);
	}
	
	/**
	 * RNS addition with accumulation, accumulating with the result of the previously executed instruction 
	 * @param b
	 * 		RNS value to add to
	 * @param destination
	 * 		destination RNS register
	 */
	public void aadd(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.AADD_VALUE, destination.name, this.name, b.name);
	}
	
	/**
	 * RNS addition with subtraction, subtracting the result of the previously executed instruction by this addition
	 * @param b
	 * 		RNS value to add to
	 * @param destination
	 * 		destination RNS register
	 */
	public void sadd(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.SADD_VALUE, destination.name, this.name, b.name);
	}
	
	/**
	 * RNS subtraction
	 * @param b
	 * 		RNS value to subtract
	 * @param destination
	 * 		destination RNS register
	 */
	public void sub(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.SUB_VALUE, destination.name, this.name, b.name);
	}
	
	/**
	 * RNS subtraction with accumulation, accumulating with the result of the previously executed instruction 
	 * @param b
	 * 		RNS value to subtract
	 * @param destination
	 * 		destination RNS register
	 */
	public void asub(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.ASUB_VALUE, destination.name, this.name, b.name);
	}

	/**
	 * RNS subtraction with subtraction, subtracting the result of the previously executed instruction by this addition
	 * @param b
	 * 		RNS value to add to
	 * @param destination
	 * 		destination RNS register
	 */
	public void ssub(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.SSUB_VALUE, destination.name, this.name, b.name);
	}
	
	/**
	 * RNS multiplication
	 * @param b
	 * 		RNS value to multiply to
	 * @param destination
	 * 		destination RNS register
	 */
	public void mult(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.MUL_VALUE, destination.name, this.name, b.name);
	}
	
	/**
	 * RNS multiplication with accumulation, accumulating with the result of the previously executed instruction 
	 * @param b
	 * 		RNS value to multiply to
	 * @param destination
	 * 		destination RNS register
	 */
	public void amult(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.AMUL_VALUE, destination.name, this.name, b.name);
	}
	
	/**
	 * RNS multiplication with subtraction, subtracting the result of the previously executed instruction by this addition
	 * @param b
	 * 		RNS value to multiply to
	 * @param destination
	 * 		destination RNS register
	 */
	public void smult(Rns b, Rns destination) {
		HW.instruction(allBases, RNSModule.SMUL_VALUE, destination.name, this.name, b.name);
	}

	
	public void modulo(Rns destination){
		modularMult(Rns.MmodN,destination);
	}
	
	
	
	/**
	 * Montgomery Special Reduction. To be used instead of modularMult(one, destination).
	 * Avoids an unnecessary RNS multiplication by one.
	 * @param destination
	 * 		destination RNS value
	 */
	public void MontSpecialReduction(Rns destination){
		
		///**
		// * 1
		// */
        //
		//HW.instruction(allBases, RNSModule.MUL_VALUE, "variable1", this.name, B.name);
		
		/**
		 * 2
		 */
		HW.instruction(allBases, RNSModule.MUL_VALUE, "variable1", this.name, "MMC2");
		
		/**
		 * 3
		 */
		HW.baseExtensionMRS(1, 2, "variable1", "variable3");
		
		/**
		 * 4
		 */
		HW.instruction(secondBase, RNSModule.MUL_VALUE, "variable3", "variable3", "MMC4");
		
		/**
		 * 5
		 */
		HW.instruction(secondBase, RNSModule.ADD_VALUE, destination.name, "variable3", "variable1");
		
		/**
		 * 6
		 */
		HW.baseExtensionMRS(2, 1, destination.name, destination.name);
		
	}
	
	
	/**
	 * Instruction that syncs both RNS processors and asks for the next bit of the key.
	 * This has no effect in the emulator, but it does affect the hardware when running the instructions.
	 */
	static public void sync(){
		HW.instruction(noBase, RNSModule.SYNC_VALUE, null, null, null);
	}
	
	/**
	 * Basic Montgomery Modular Multiplication ( MM ) with both MRS base extensions 
	 * @param B
	 * 		RNS value to multiply to
	 * @param destination
	 * 		destination RNS value
	 */
	public void modularMult(Rns B,Rns destination){
		
		/**
		 * 1
		 */

		HW.instruction(allBases, RNSModule.MUL_VALUE, "variable1", this.name, B.name);
		/**
		 * 2
		 */
		HW.instruction(allBases, RNSModule.MUL_VALUE, "variable1", "variable1", "MMC2");
		
		/**
		 * 3
		 */
		HW.baseExtensionMRS(1, 2, "variable1", "variable3");
		
		/**
		 * 4
		 */
		HW.instruction(secondBase, RNSModule.MUL_VALUE, "variable3", "variable3", "MMC4");
		
		/**
		 * 5
		 */
		HW.instruction(secondBase, RNSModule.ADD_VALUE, destination.name, "variable3", "variable1");
		
		/**
		 * 6
		 */
		HW.baseExtensionMRS(2, 1, destination.name, destination.name);
		
	}
	
	


	
	/**
	 * Convert a number represented in the different hardware channel registers of the first base to binary 
	 * @param name
	 *		register to convert
	 * @return BigInteger with the binary value
	 */
	public static BigInteger RNStoIntB1(String name) {
		BigInteger M1 = new BigInteger("1");
		BigInteger[] Mi1 = new BigInteger[h1];
		BigInteger[] MiToMinus1 =  new BigInteger[h1];
		BigInteger[] Ei = new BigInteger[h1];
		BigInteger[] aux = new BigInteger[h1];
		BigInteger value= new BigInteger("0");
		
		/* Constants-------------------------- */
		for (int i = 0; i < Rns.h1; i++) {
			M1 = M1.multiply(base1mod[i]);
		}

		for (int i = 0; i < Rns.h1; i++) {
			Mi1[i] = M1.divide(base1mod[i]);
		}
		for (int i = 0; i < Rns.h1; i++) {
			MiToMinus1[i]= Mi1[i].modInverse(base1mod[i]);
		}
		/*------------------------------------*/
	
		for (int i = 0; i < Rns.h1; i++) {
			Ei[i] = Hardware.HWbase1mod[i].registerBank.get(name).multiply(MiToMinus1[i]); //Hardware.HWbase1mod[i].R("variable1");
		}

		for (int i = 0; i < Rns.h1; i++) {
			aux[i] = Ei[i].multiply(Mi1[i]);
			value = value.add(aux[i]).mod(M1);
		}
		return value;
		
	}
	
	/**
	 * Convert a number represented in the different hardware channel registers of the second base to binary 
	 * @param name
	 * 		register to convert
	 * @return BigInteger with the binary value
	 */
	public static BigInteger RNStoIntB2(String name) {
		BigInteger M2 = new BigInteger("1");
		BigInteger[] Mi2 = new BigInteger[h2];
		BigInteger[] MiToMinus1 =  new BigInteger[h2];
		BigInteger[] Ei = new BigInteger[h1];
		BigInteger[] aux = new BigInteger[h1];
		BigInteger value= new BigInteger("0");
		
		/* Constants-------------------------- */
		for (int i = 0; i < Rns.h2; i++) {
			M2 = M2.multiply(base2mod[i]);
		}

		for (int i = 0; i < Rns.h2; i++) {
			Mi2[i] = M2.divide(base2mod[i]);
		}
		for (int i = 0; i < Rns.h2; i++) {
			MiToMinus1[i]= Mi2[i].modInverse(base2mod[i]);
		}
		/*------------------------------------*/
	
		for (int i = 0; i < Rns.h2; i++) {
			Ei[i] = Hardware.HWbase2mod[i].registerBank.get(name).multiply(MiToMinus1[i]); //Hardware.HWbase1mod[i].R("variable1");
		}

		for (int i = 0; i < Rns.h1; i++) {
			aux[i] = Ei[i].multiply(Mi2[i]);
			value = value.add(aux[i]).mod(M2);
		}
		return value;
		
	}
	
	/**
	 * Convert this RNS value to binary (uses the first base), via MRC.
	 * Prints converted result to the output file.
	 * @return BigInteger with the binary value
	 */
	public BigInteger toIntMRC() { 
		BigInteger result=toIntMRC_simul();
		
		Hardware.printHexn(result,Main.BITS_PER_MOD,Main.resultFile);
		Main.assemblyFile.println("-- "+"01" + " " + "RNS-to-MRC" + " " + "aux" + " " + this.name + " "+ "-----");
		Hardware.printHexInstruction("01", "0110", "0", "aux", this.name, null, Main.codeFile);
		Main.assemblyFile.println("-- "+"01" + " " + "MRC-to-Bin" + " " + "-----"+ " " + "aux" + " "+ "-----");
		Hardware.printHexInstruction("01", "0111", "0", null, "aux", null, Main.codeFile);
		
		return result;
	}
	
	
	/**
	 * Convert this RNS value to binary (uses the first base), via MRC
	 * @return BigInteger with the binary value
	 */
	public BigInteger toIntMRC_simul() {
		/* MRC */
		BigInteger aux;
		BigInteger value;
		
		
		
		
		
		for (int j = 0; j < h1; j++) {
			if (j == 0) {
				HW.instructionNoPrint(firstBase, RNSModule.MUL_VALUE, "variable1", this.name, "newMRCC" + j);
			} else {
				HW.instructionNoPrint(firstBase, RNSModule.AMUL_VALUE, "variable1", "OUTBUS", "newMRCC" + j);
			}
			Hardware.HWbase1mod[j].writeOutRegister("variable1");
		}

		value = BigInteger.ZERO;
		for(int i=0;i<h1;i++){
			aux=BigInteger.ZERO;
			HW.instructionNoPrint(firstBase, RNSModule.MUL_VALUE, "variable2", "variable1", "Wij" +i);
			for(int j=0;j<h1;j++){
				aux=aux.add(Hardware.HWbase1mod[j].hardwareBlock.bin_product);
			}
			aux=aux.shiftLeft(Main.BITS_PER_MOD*i);
			value=value.add(aux);
		}

		return value;

	}

	/**
	 * RNS modular exponentiation
	 * @param exp
	 * 				binary power to apply
	 * @return
	 * 				Result of the exponentiation, in RNS
	 */
	public Rns exp(BigInteger exp) {
		Rns x1 = new Rns("x1");
		Rns x2 = new Rns("x2");
		x1 = this;
		this.modularMult(this,  x2);
		for (int i = exp.bitLength() - 2; i >= 0; i--) {
			if (exp.testBit(i) == false) {
				x1.modularMult(x2, x2);
				x1.modularMult(x1, x1);
				sync();
			} else {
				x1.modularMult(x2, x1);
				x2.modularMult(x2, x2);
				sync();
			}
		}

		return x1;

	}

	
	
	/**
	 * Must be called first, before any other use of RNS. Configures RNS channels including hardware.
	 * @param N
	 *            cryptographic modulo
	 */
	public static void computeModuli(BigInteger N) {
		BigInteger[] m = new BigInteger[2 * Main.MODULI_NO];
		BigInteger one = new BigInteger("1");
		BigInteger two = new BigInteger("2");
		BigInteger M1 = one;
		BigInteger M2 = one;

		
		
		Rns.n = Main.BITS_PER_MOD;
		Rns.h1 = Main.MODULI_NO;
		Rns.h2 = h1;
		Rns.modulo = N;

		System.out.println(n);

		Rns.base1mod = new BigInteger[h1];
		Rns.base2mod = new BigInteger[h2];
		HW = new Hardware();

		BigInteger current_moduli;

		current_moduli = two.pow(n);
		System.out.println(current_moduli);
		current_moduli = current_moduli.subtract(one);

		for (int i = 0; i < 2 * h1; i++) {
			while (true) {
				boolean passes_all_tests = true;

				for (int j = 0; j < i; j++) {
					if (current_moduli.gcd(m[j]).compareTo(one) != 0) {
						passes_all_tests = false;
						break;
					}
				}

				if (passes_all_tests) {
					break;
				} else {
					current_moduli = current_moduli.subtract(two);
				}

			}

			m[i] = current_moduli;
			current_moduli = current_moduli.subtract(two);

			if (two.pow(n).subtract(one).subtract(current_moduli).bitLength() > n / 2) {
				System.out.println("Constraint not met, aborting: width of k <= n/2");
				System.exit(-1);
			}

		}

		for (int i = 0; i < h1; i++) {
			Rns.base1mod[i] = m[2 * i + 1];
			HW.setChannel(m[2 * i + 1], i, 1);
			Rns.base2mod[i] = m[2 * i];
			HW.setChannel(m[2 * i], i, 2);

			System.out.println(Rns.base1mod[i] + "(k=" + two.pow(n).subtract(Rns.base1mod[i]) + ",w_ki="+two.pow(n).subtract(Rns.base1mod[i]).bitLength()+")   " + Rns.base2mod[i] + "(k="
					+ two.pow(n).subtract(Rns.base2mod[i]) + ",w_ki="+two.pow(n).subtract(Rns.base1mod[i]).bitLength()+ ")");
			M1 = M1.multiply(Rns.base1mod[i]);
			M2 = M2.multiply(Rns.base2mod[i]);
		}

		System.out.println(M1 + "   " + M2);
		
		if(N.compareTo(new BigInteger("2").pow(Main.KEY_SIZE).subtract(BigInteger.ONE)) == 0 ){
			if(M1.compareTo(modulo.multiply(new BigInteger("4"))) <= 0 || M1.compareTo(M2)>0 ){
				System.out.println("Condition not verified: M1 > 4N or M1 < M2. Increasing number of moduli and retrying...");
				Main.MODULI_NO++;
				computeModuli(N);
			}
		}else{
			if(M1.compareTo(modulo.multiply(new BigInteger("4"))) <= 0 || M1.compareTo(M2)>0 || M1.gcd(modulo).compareTo(BigInteger.ONE)!=0 || M2.gcd(modulo).compareTo(BigInteger.ONE)!=0 ){
				System.out.println("Condition not verified: M1 > 4N or M1 < M2 or GCD(M_i,N)=1. Increasing number of moduli and retrying...");
				Main.MODULI_NO++;
				computeModuli(N);
			}
		}

//		Main.printModuli();

	}
	
	

	/**
	 * Verifies if Montgomery Multiplication Constraints are met.
	 */
	public static void verifyConstraints(){
		BigInteger M1 = new BigInteger("1");
		for(int i=0; i < Rns.h1 ; i++){
			M1 = M1.multiply(Rns.base1mod[i]);
		}
		BigInteger M2 = new BigInteger("1");
		for(int i=0; i < Rns.h2 ; i++){
			M2 = M2.multiply(Rns.base2mod[i]);
		}
		
		if(M1.compareTo(modulo.multiply(new BigInteger("4"))) <= 0 || M1.compareTo(M2)>0 ){
			System.out.println("/!\\/!\\/!\\/!\\ ALERT!!! /!\\/!\\/!\\/!\\ ");
			System.out.println("Condition not verified: M1 > 4N or M1 > M2");
			System.out.println("This RNS configuration cannot be used with the desired modulo");
			System.exit(-1);
//			Main.MODULI_NO++;
//			Rns.computeModuli(modulo);
//			return false;
		}
		if(M1.gcd(modulo).compareTo(BigInteger.ONE)!=0 || M2.gcd(modulo).compareTo(BigInteger.ONE)!=0){
			System.out.println("This RNS configuration cannot be used with the desired modulo");
			System.out.println("M1 and M2 must be coprime with N, i.e. this is happening GCD(M_i,N)!=1");
			System.exit(-1);
		}
	}
	
	
	/**
	 * Generates RNS constants. Computes the constants required for the other RNS methods
	 */
	static void generateConstants() {
		/**
		 * Forward Conversion
		 */
		
		/* BASE 1 */
		for (int i = 0; i < h1; i++) {
			for (int j = 1; j <= Math.ceil(Main.KEY_SIZE / (float) n); j++) {
				Hardware.HWbase1mod[i].addRegister("2^(n*j-1)" + j, new BigInteger("2").pow(n * (j - 1)).mod(base1mod[i]));
			}
			Hardware.HWbase1mod[i].addRegister("OUTBUS", BigInteger.ZERO);
		}

		/* BASE 2 */
		for (int i = 0; i < h2; i++) {
			for (int j = 1; j <= Math.ceil(Main.KEY_SIZE / (float) n); j++) {
				Hardware.HWbase2mod[i].addRegister("2^(n*j-1)" + j, new BigInteger("2").pow(n * (j - 1)).mod(base2mod[i]));
			}
			Hardware.HWbase2mod[i].addRegister("OUTBUS", BigInteger.ZERO);
		}


		BigInteger M1 = new BigInteger("1");
		BigInteger M2 = new BigInteger("1");
		for (int i = 0; i < Rns.h1; i++) {
			M1 = M1.multiply(base1mod[i]);
			M2 = M2.multiply(base2mod[i]);
		}

		

		/**MRC----------------------------------------------------*/
		BigInteger[][] mTimesMmodInverseBase1 = new BigInteger[h1][h1];
		BigInteger aux;
		for (int i = 0; i < h1; i++) {
			for (int j = 0; j < h1; j++) {
				mTimesMmodInverseBase1[i][j] = BigInteger.ZERO;
			}
		}
		mTimesMmodInverseBase1[0][0] = BigInteger.ONE;
		for (int i = 1; i < h1 ; i++) {
			for (int j = 1; j <= i; j++) {
				if (j == 1) {
					aux = BigInteger.ONE;
					for (int k = i; k > 0; k--) {
						aux = base1mod[k - 1].modInverse(base1mod[i]).multiply(aux).mod(base1mod[i]);
					}
					mTimesMmodInverseBase1[i][j - 1] = aux;
					mTimesMmodInverseBase1[i][j] = aux;
				}else{
					mTimesMmodInverseBase1[i][j]=mTimesMmodInverseBase1[i][j-1].multiply(base1mod[j-2]).mod(base1mod[i]);
				}
			}
		}
		for (int i = 0; i < h1; i++) {
			Hardware.HWbase1mod[i].addRegister("newMRCC"+0, mTimesMmodInverseBase1[i][0]);
			for (int j = 1; j < h2; j++) {
				Hardware.HWbase1mod[i].addRegister("newMRCC"+j, BigInteger.ZERO.subtract(mTimesMmodInverseBase1[i][j]).mod(base1mod[i]));
			}
		}
		
		
		/**reverse conversion*/
		BigInteger[] mTimesM = new BigInteger[h1];
		mTimesM[0] = BigInteger.ONE;
		for (int i = 1; i < h1; i++) {
			mTimesM[i] = mTimesM[i - 1].multiply(base1mod[i - 1]);
		}

		for(int i = 0 ; i<h1;i++){
			for(int j=0;j<h1;j++){
				Hardware.HWbase1mod[i].addRegister("Wij"+j, mTimesM[i].shiftRight(Main.BITS_PER_MOD*j).and(new BigInteger("2").pow(Main.BITS_PER_MOD).subtract(BigInteger.ONE)));
			}
		}
		
		
		/**BE*/
		
//		BigInteger[][] mTimesMmodInverseBase1 = new BigInteger[Rns.h1][Rns.h1];
		BigInteger[][] mTimesMmodInverseBase2 = new BigInteger[Rns.h1][Rns.h1];
		BigInteger aux2;
		
		
		
		for (int i = 0; i < Rns.h1; i++) {
			for (int j = 0; j < Rns.h1; j++) {
				mTimesMmodInverseBase1[i][j] = BigInteger.ZERO;
				mTimesMmodInverseBase2[i][j] = BigInteger.ZERO;
			}
		}
		mTimesMmodInverseBase1[0][0] = BigInteger.ONE;
		mTimesMmodInverseBase2[0][0] = BigInteger.ONE;
		for (int i = 1; i < Rns.h1 ; i++) {
			for (int j = 1; j <= i; j++) {
				if (j == 1) {
					aux = BigInteger.ONE;
					aux2 = BigInteger.ONE;
					for (int k = i; k > 0; k--) {
						aux = Rns.base1mod[k - 1].modInverse(Rns.base1mod[i]).multiply(aux).mod(Rns.base1mod[i]);
						aux2 = Rns.base2mod[k - 1].modInverse(Rns.base2mod[i]).multiply(aux2).mod(Rns.base2mod[i]);
					}
					mTimesMmodInverseBase1[i][j - 1] = aux;
					mTimesMmodInverseBase1[i][j] = aux;
					mTimesMmodInverseBase2[i][j - 1] = aux2;
					mTimesMmodInverseBase2[i][j] = aux2;
				}else{
					mTimesMmodInverseBase1[i][j]=mTimesMmodInverseBase1[i][j-1].multiply(Rns.base1mod[j-2]).mod(Rns.base1mod[i]);
					mTimesMmodInverseBase2[i][j]=mTimesMmodInverseBase2[i][j-1].multiply(Rns.base2mod[j-2]).mod(Rns.base2mod[i]);
				}
			}
		}
		for (int i = 0; i < Rns.h1; i++) {
			Hardware.HWbase1mod[i].addRegister("12mrsBE"+0, mTimesMmodInverseBase1[i][0]);
			Hardware.HWbase2mod[i].addRegister("21mrsBE"+0, mTimesMmodInverseBase2[i][0]);
			for (int j = 1; j < Rns.h2; j++) {
				Hardware.HWbase1mod[i].addRegister("12mrsBE"+j, BigInteger.ZERO.subtract(mTimesMmodInverseBase1[i][j]).mod(Rns.base1mod[i]));
				Hardware.HWbase2mod[i].addRegister("21mrsBE"+j, BigInteger.ZERO.subtract(mTimesMmodInverseBase2[i][j]).mod(Rns.base2mod[i]));
			}
			Hardware.HWbase1mod[i].addRegister("12mrsBE"+Rns.h2, BigInteger.ONE);
			Hardware.HWbase2mod[i].addRegister("21mrsBE"+Rns.h2, BigInteger.ONE);
		}
		BigInteger[][] mTimesMmodBase2 = new BigInteger[Rns.h1][Rns.h1];
		BigInteger[][] mTimesMmodBase1 = new BigInteger[Rns.h1][Rns.h1];
		for (int i = 0; i < Rns.h1; i++) {
			mTimesMmodBase2[i][0] = BigInteger.ONE;
			mTimesMmodBase1[i][0] = BigInteger.ONE;
			for (int j = 0; j < Rns.h1 - 1; j++) {
				mTimesMmodBase2[i][j + 1] = Rns.base1mod[j].multiply(mTimesMmodBase2[i][j]).mod(Rns.base2mod[i]);
				mTimesMmodBase1[i][j + 1] = Rns.base2mod[j].multiply(mTimesMmodBase1[i][j]).mod(Rns.base1mod[i]);
			}
		}
		
		for (int i = 0; i < Rns.h1; i++) {
			Hardware.HWbase2mod[i].addRegister("12mrsBE"+0,BigInteger.ZERO);
			Hardware.HWbase1mod[i].addRegister("21mrsBE"+0,BigInteger.ZERO);
			for (int j = 0; j < Rns.h1; j++) {
				Hardware.HWbase2mod[i].addRegister("12mrsBE" + (j+1), mTimesMmodBase2[i][j].mod(Rns.base2mod[i]));
				Hardware.HWbase1mod[i].addRegister("21mrsBE" + (j+1), mTimesMmodBase1[i][j].mod(Rns.base1mod[i]));
			}
		}
		
	}
	
	/**
	 * Generates Montgomery Modular Multiplication specific constants configuring a specific modulo.
	 * Verifies if provided modulo is suitable for the hardware architecture.
	 * @param N BigInteger with the value of the modulo
	 */
	public static void generateMMConstantsFor(BigInteger N){
		Rns.modulo=N;
		
		verifyConstraints();
		
		if(Main.dataFile == null || Main.codeFile == null || Main.moduliFile == null || Main.resultFile == null){
			System.out.println("Main.createFiles() must be called before generating constants. Aborting...");
			System.exit(-1);
		}
		BigInteger M1 = new BigInteger("1");
		for(int i=0; i < Rns.h1 ; i++){
			M1 = M1.multiply(Rns.base1mod[i]);
		}
		BigInteger M2 = new BigInteger("1");
		for(int i=0; i < Rns.h2 ; i++){
			M2 = M2.multiply(Rns.base2mod[i]);
		}
		
		
		/*regular MM*/
		for (int i = 0; i < h1; i++) {
			Hardware.HWbase1mod[i].addRegister("MMC2",BigInteger.ZERO.subtract(Rns.modulo.mod(base1mod[i]).modInverse(base1mod[i])).mod(base1mod[i]));
			Hardware.HWbase2mod[i].addRegister(
					"MMC2",
					M1.modInverse(base2mod[i]).mod(base2mod[i]));
			Hardware.HWbase2mod[i].addRegister("MMC4",Rns.modulo.mod(base2mod[i]).multiply(M1.modInverse(base2mod[i])).mod(base2mod[i]));
		}
		
		
		one=new Rns("1","one"); /*to convert from the MD*/
		MMmodN=new Rns(M1.multiply(M1).mod(modulo).toString(),"M^2modN"); /*to convert to the MD*/
		MmodN=new Rns(M1.mod(modulo).toString(),"MmodN"); /* to perform modulo operation */
		
		Hardware.printHexn(Rns.RNStoIntB2("MMC4"),Main.BITS_PER_MOD,Main.dataFile);
		Hardware.printHexn(Rns.RNStoIntB2("MMC2"),Main.BITS_PER_MOD,Main.dataFile);
		Hardware.printHexn(Rns.RNStoIntB1("MMC2"),Main.BITS_PER_MOD,Main.dataFile);
		
		Main.assemblyFile.println("-- "+"01" + " " + "Bin-to-RNS_p" + " " + "MMC4" + " " + "-----" + " "+ "-----");
		Hardware.printHexInstruction("01","0100","1","MMC4",null,null, Main.codeFile);
		Main.assemblyFile.println("-- "+"10" + " " + "Bin-to-RNS" + " " + "MMC4" + " " + "-----" + " "+ "-----");
		Hardware.printHexInstruction("10","0100","0","MMC4",null,null, Main.codeFile);
		
		Main.assemblyFile.println("-- "+"01" + " " + "Bin-to-RNS_p" + " " + "MMC2" + " " + "-----" + " "+ "-----");
		Hardware.printHexInstruction("01","0100","1","MMC2",null,null, Main.codeFile);
		Main.assemblyFile.println("-- "+"10" + " " + "Bin-to-RNS" + " " + "MMC2" + " " + "-----" + " "+ "-----");
		Hardware.printHexInstruction("10","0100","0","MMC2",null,null, Main.codeFile);
		
		Main.assemblyFile.println("-- "+"01" + " " + "Bin-to-RNS" + " " + "MMC2" + " " + "-----" + " "+ "-----");
		Hardware.printHexInstruction("01","0100","0","MMC2",null,null, Main.codeFile);
		
		
	}
	
	
	/**
	 * Converts this Rns value to the Montgomery Domain. Appends _MD to the name of the register and returns it.
	 * @return
	 * 		the value in the Montgomery Domain.
	 */
	public Rns toMD(){
		Rns result = new Rns(name+"_MD");
		modularMult(MMmodN, result);
		return result;
	}
	
	/**
	 * Converts this Rns value from the Montgomery Domain. Appends _OD to the name of the register and returns it.
	 * @return
	 * 		the value in the Original Domain.
	 */
	public Rns fromMD(){
		Rns result = new Rns(name+"_OD");
		MontSpecialReduction(result);
		return result;
	}



	
}
