package examples;

import java.math.BigInteger;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import rns.ECC;
import rns.ECPointRNS;
import rns.Main;
import rns.Rns;

public class Example {

	/*----------------------- ATTENTION !! ------------------------
	|       This is just a copy of the source file located in      |
	|        \Eclipse Project and Source Code\rns\src\examples .   |
	|     Changes in this file will produce no results whatsoever. |
	|    If you wish to modify it or write your own, please use    |
	|              the class in the project directory              |
	--------------------------------------------------------------*/	
	static void modularMultExample(){
				
		/*******************************************************************************************************/
		/** Define modulo and hardware           											                  **/
		/*******************************************************************************************************/
		
		/* Set the modulo for the modulo operation */
		BigInteger N = new BigInteger("FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFEFFFFFFFFFFFFFFFF",16);  /* modulo */
		
		
		/*Define number of bits of the operands and the number of bits in the RNS architecture.
		 *Creates the ModuliSet and hardware structure suitable for Modular operation. 
		 *Prints hardware configuration to mmult.xml. */
		Main.createModuliSetWithMM(256, 16, "mmult.xml",N);
		
		
		/*Alternatively, moduli set and hardware structure can be loaded from xml file*/
//		Main.loadModuliSet("mmult.xml");
	
		/* Creates the output files, appending MMULT to their names 
		 * Must be called after creating or loading ModuliSet*/
		Main.createFiles("MMULT");
		
		/* Generate and write Montgomery Modular Multiplication constants.
		 * Verifies if the current moduliSet is compatible with the selected modulo N*/
		Rns.generateMMConstantsFor(N);
		/*******************************************************************************************************/
		
		
		
		/*******************************************************************************************************/
		/** Modular Multiplication Example										                              **/
		/*******************************************************************************************************/
		
		/* Operands */
		BigInteger A_bin = new BigInteger("FFFFFFFFFFFFFFFFFFFF0000FFFFFFFEFFFFFFFFFFFEFFFE",16);
		BigInteger B_bin = new BigInteger("FFFFFFFFFFFFFFFFFFFFFFFFFFF0000EFFFFFFFFFFFFFFFE",16);
		BigInteger C_bin;
		Rns A = new Rns(A_bin.toString(),"A");      //values are converted to RNS 
		Rns B = new Rns(B_bin.toString(),"B");      //values are converted to RNS 
		Rns C = new Rns("C");
		Rns A_MD = new Rns("A_MD");  /* register allocation to be used in the Montgomery Domain representation */
		Rns B_MD = new Rns("B_MD");  /* register allocation to be used in the Montgomery Domain representation */
		Rns C_MD = new Rns("C_MD");  /* register allocation to be used in the Montgomery Domain representation */
		
		/* Convert operands to the Montgomery Domain */
		A_MD = A.toMD();
		B_MD = B.toMD();
		
		/* C = A x B mod N */
		A_MD.modularMult(B_MD, C_MD);
		
		/* Convert result from the Montgomery Domain */
		C = C_MD.fromMD();
		
		/* Convert back to binary*/
		C_bin = C.toIntMRC();
		
		/*******************************************************************************************************/
		
		
		
		
		/* Closes the output files */
		Main.closeFiles();
		
		System.out.println(C_bin + " which should be equal to \n" + A_bin.multiply(B_bin).mod(N));
	}
	
}
