﻿package examples;

import java.math.BigInteger;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;

import rns.ECC;
import rns.ECPointRNS;
import rns.Main;
import rns.Rns;

public class Example {

	/*----------------------- ATTENTION !! ------------------------
	|       This is just a copy of the source file located in      |
	|        \Eclipse Project and Source Code\rns\src\examples .   |
	|     Changes in this file will produce no results whatsoever. |
	|    If you wish to modify it or write your own, please use    |
	|              the class in the project directory              |
	--------------------------------------------------------------*/
	
static void ECPointAddition(){
		
		
		/*******************************************************************************************************/
		/** Define modulo and hardware           											                  **/
		/*******************************************************************************************************/
		
		/* Set the modulo for the modulo operation and EC parameters */
		BigInteger p = new BigInteger("263");   /* modulo */
		BigInteger a_bin = new BigInteger("2");   /* Curve parameters */
		BigInteger b_bin = new BigInteger("3");   /* Curve parameters */
		
		
		/*Define number of bits of the operands and the number of bits in the RNS architecture.
		 *Creates the ModuliSet and hardware structure suitable for Modular operation. 
		 *Prints hardware configuration to padd.xml. */
		Main.createModuliSetWithMM(256, 16, "padd.xml",p);
		
		
		/*Alternatively, moduli set and hardware structure can be loaded from xml file*/
//		Main.loadModuliSet("padd.xml");

		/* Creates the output files, appending MMULT to their names 
		 * Must be called after creating or loading ModuliSet*/
		Main.createFiles("PADD");
		
		/* Generate and write Montgomery Modular Multiplication EC constants.
		 * Verifies if the current moduliSet is compatible with the selected modulo p*/
		ECPointRNS.generateECconstants(a_bin,b_bin,p);
		/*******************************************************************************************************/
		
		
		
		
		
		/*******************************************************************************************************/
		/** Elliptic Curve Point Addition										                              **/
		/*******************************************************************************************************/
		
		
		/* Set the binary inputs */
		BigInteger X1_bin = new BigInteger("14");
		BigInteger Y1_bin = new BigInteger("113");

		BigInteger X2_bin = new BigInteger("13");
		BigInteger Y2_bin = new BigInteger("139");
		
		BigInteger X3_bin;
		BigInteger Y3_bin;
		BigInteger Z3_bin;

		

		
		Main.codeFile.println("-- Point Add Start");
		Main.dataFile.println("-- Point Add Start");
		
		/* Operands for P3 = P1 + P2*/ /* P(x,y) */
		
		/*Inputs*/
		Rns X1      = new Rns(X1_bin.toString(),"X1");
		Rns Y1      = new Rns(Y1_bin.toString(),"Y1");
		
		Rns X2      = new Rns(X2_bin.toString(),"X2");
		Rns Y2      = new Rns(Y2_bin.toString(),"Y2");
		
		/*Outputs - projective coordinates of P3*/
		Rns X3;
		Rns Y3;
		Rns Z3;
		
		
		
		/*Computation begins*/		
		/*Intermediate values*/
		Rns X1_MD;
		Rns Y1_MD;
		Rns X2_MD;
		Rns Y2_MD;
		
		Rns Z1_MD      = new Rns("Z1_MD");
		Rns Z2_MD      = new Rns("Z2_MD");
		Rns X3_MD = new Rns("X3_MD");
		Rns Y3_MD = new Rns("Y3_MD");
		Rns Z3_MD = new Rns("Z3_MD");
		Rns A       = new Rns("A");
		Rns B       = new Rns("B");
		Rns C       = new Rns("C");
		Rns D       = new Rns("D");
		Rns E       = new Rns("E");
		Rns F       = new Rns("F");
		Rns G       = new Rns("G");
		Rns H       = new Rns("H");
		Rns I       = new Rns("I");
		Rns lambdaN = new Rns("lambdaN");
		Rns lambdaD = new Rns("lambdaD");
		
		
		/* Convert operands to the Montgomery Domain */
		X1_MD = X1.toMD();
		Y1_MD = Y1.toMD();
		X2_MD =  X2.toMD();
		Y2_MD =  Y2.toMD();
		
		Rns.one.mult(Rns.MmodN, Z1_MD);     /* initialize Z projective coordinate to 1, in the Montgomery Domain. Z1 = 1 . M mod p)*/
		Rns.one.mult(Rns.MmodN, Z2_MD);     /* initialize Z projective coordinate to 1, in the Montgomery Domain. Z2 = 1 . M mod p)*/
		
		/* Perform point addition according to unified addition formulae in */
		/* Jean-Claude Bajard, Sylvain Duquesne, Milos D. Ercegovac. COMBINING LEAK–
		 * RESISTANT ARITHMETIC FOR ELLIPTIC CURVES DEFINED OVER F p AND RNS
		 * REPRESENTATION. Publications Mathematiques UFR Sciences Techniques Besan¸ con, 2013,
		 * pp.67-87
		 */
		
		/** A = X2.Z1 */
		X2_MD.modularMult(Z1_MD, A);
		
		/** B = X1.Z2 */
		X1_MD.modularMult(Z2_MD, B);
		
		/** C = Z1.Z2 */
		Z1_MD.modularMult(Z2_MD, C);
		
		/** D = a.C */
		ECPointRNS.a_MD.modularMult(C, D);
		
		/** lambdaN = (A + B)^2 - A.B + C.D */
		/*I=A+B*/
		A.add(B, I);
		I.modulo(I);
		
		I.modularMult(I, lambdaN);
		A.modularMult(B, B);
		C.modularMult(D, D);
		
		lambdaN.sub(B, lambdaN);
		lambdaN.modulo(lambdaN);
		lambdaN.add(D, lambdaN);
		lambdaN.modulo(lambdaN);
		
		/** E = Y1.Z2 + Y2.Z1 */
		Y1_MD.modularMult(Z2_MD, E);
		Y2_MD.modularMult(Z1_MD, H);
		E.add(H, E);
		E.modulo(E);
		
		
		/** lambdaD = E.C */
		E.modularMult(C, lambdaD);
		
		/**F = E.lambdaD */
		E.modularMult(lambdaD, F);
		
		/** G = lambdaN.lambdaN */
		lambdaN.modularMult(lambdaN, G);
		
		/** H = F.(A+B) = F.I */
		F.modularMult(I, H);
		
		/** X3 = 2.lambdaD . (G - H) */
		G.sub(H, A);
		A.modulo(A);
		
		A.modularMult(lambdaD, X3_MD);
		X3_MD.mult(ECPointRNS.two, X3_MD);
		X3_MD.modulo(X3_MD);
		
		/** Y3 = lambdaN(3.H - 2.G) - F^2 */
		F.modularMult(F, F);
		G.mult(ECPointRNS.two, G);
		G.modulo(G);
		ECPointRNS.two.add(Rns.one, B); //3
		B.mult(H, H);
		H.modulo(H);
		
		H.sub(G, H);
		H.modulo(H);
		lambdaN.modularMult(H, lambdaN);
		lambdaN.sub(F, Y3_MD);
		Y3_MD.modulo(Y3_MD);
		
		/** Z3 = 2.lambdaD^3 */
		lambdaD.modularMult(lambdaD, A);
		A.modularMult(lambdaD, Z3_MD);
		Z3_MD.mult(ECPointRNS.two, Z3_MD);
		Z3_MD.modulo(Z3_MD);
		
			
		/* Convert result from the Montgomery Domain */
//		Main.codeFile.println("-- Conversion to OD");
		X3 = X3_MD.fromMD();
		Y3 = Y3_MD.fromMD();
		Z3 = Z3_MD.fromMD();
		
		/* Convert back to binary*/
//		Main.codeFile.println("-- Conversion to bin");
		X3_bin = X3.toIntMRC();
		Y3_bin = Y3.toIntMRC();
		Z3_bin = Z3.toIntMRC();
		
		/*******************************************************************************************************/
		
		
		
		/* Closes the output files */
		Main.closeFiles();
		
		/*Binary verification of the result*/
		System.out.println("x = " + X3_bin.multiply(Z3_bin.modInverse(p)).mod(p));
		System.out.println("y = " + Y3_bin.multiply(Z3_bin.modInverse(p)).mod(p));
		
		
		
		
		/**Binary Version*/
/*		BigInteger a = new BigInteger("2");  
		BigInteger b = new BigInteger("3"); 
		BigInteger p = new BigInteger("263"); 
		
		BigInteger X1 = new BigInteger("14");
		BigInteger Y1 = new BigInteger("113");
		BigInteger Z1 = new BigInteger("1");
		
		BigInteger X2 = new BigInteger("13");
		BigInteger Y2 = new BigInteger("139");
		BigInteger Z2 = new BigInteger("1");
		
		BigInteger X3;
		BigInteger Y3;
		BigInteger Z3;
		
		BigInteger lambdaN;
		BigInteger lambdaD;
		BigInteger A;
		BigInteger B;
		BigInteger C;
		BigInteger D;
		BigInteger E;
		BigInteger F;
		BigInteger G;
		BigInteger H;
//		
//		temp
//		lambdaN = X1.multiply(Z2).add(X2.multiply(Z1)).pow(2).subtract(X1.multiply(X2).multiply(Z1.multiply(Z2)).add(a.multiply(Z1.pow(2).multiply(Z2.pow(2))))).mod(p);
//		lambdaD = Y1.multiply(Z2).add(Y2.multiply(Z1)).multiply(Z1).multiply(Z2).mod(p);
//		
//		
//		X3 = new BigInteger("2").multiply(lambdaD).multiply(lambdaN.pow(2).subtract(X1.multiply(X2).add(X2.multiply(Z1)).multiply(Y1.multiply(Z2).add(Y2.multiply(Z1))).multiply(lambdaD))).mod(p);
		
		A = X2.multiply(Z1).mod(p);
		B = X1.multiply(Z2).mod(p);
		C = Z1.multiply(Z2).mod(p);
		D = a.multiply(C).mod(p);
		lambdaN = A.add(B).pow(2).subtract(A.multiply(B)).add(C.multiply(D)).mod(p);
		
		E = Y1.multiply(Z2).add(Y2.multiply(Z1)).mod(p);
		lambdaD = E.multiply(C).mod(p);
		
		F = E.multiply(lambdaD).mod(p);
		G = lambdaN.pow(2).mod(p);
		H = F.multiply(A.add(B)).mod(p);
		
		X3 = new BigInteger("2").multiply(lambdaD).multiply(G.subtract(H)).mod(p);
		Y3 = lambdaN.multiply(new BigInteger("3").multiply(H).subtract(new BigInteger("2").multiply(G))).subtract(F.pow(2)).mod(p);
		Z3 = new BigInteger("2").multiply(lambdaD.pow(3)).mod(p);
		
		System.out.println("x = " + X3.multiply(Z3.modInverse(p)).mod(p));
		System.out.println("y = " + Y3.multiply(Z3.modInverse(p)).mod(p));
		
		System.out.println("A = " + A);
		System.out.println("B = " + B);
		System.out.println("C = " + C);
		System.out.println("D = " + D);
		System.out.println("E = " + E);
		System.out.println("F = " + F);
		System.out.println("G = " + G);
		System.out.println("H = " + H);
		System.out.println("lambdaN = " + lambdaN);
		System.out.println("lambdaD = " + lambdaD);
		System.out.println("X3 = " + X3);
		System.out.println("Y3 = " + Y3);
		System.out.println("Z3 = " + Z3); */
	}



	
}
